# Utility functions to make testing easier
import asyncio
from typing import Any, Awaitable


def run_coroutine(result: Awaitable[Any]) -> Any:
    """Wrapper for asyncio functions to allow them to be run from synchronous functions"""
    loop = asyncio.get_event_loop()
    result = loop.run_until_complete(result)
    loop.close()
    return result


def make_awaitable(result: Any = None) -> Awaitable[Any]:
    """
    Makes an awaitable, suitable for mocking an `async` function.
    This uses Futures as they can be awaited multiple times so can be returned
    to multiple callers.
    """
    future = asyncio.Future()  # type: ignore
    future.set_result(result)
    return future



import logging
import os
import re
import sys
from typing import Any, List, Optional

import pytimeparse2
import yaml

from matrix_alertbot.errors import (
    InvalidConfigError,
    ParseConfigError,
    RequiredConfigKeyError,
)

logger = logging.getLogger()
logging.getLogger("peewee").setLevel(
    logging.INFO
)  # Prevent debug messages from peewee lib


DEFAULT_REACTIONS = {"🤫", "😶", "🤐", "🙊", "🔇", "🔕"}


class Config:
    """Creates a Config object from a YAML-encoded config file from a given filepath"""

    def __init__(self, filepath: str):
        self.filepath = filepath
        if not os.path.isfile(filepath):
            raise ParseConfigError(f"Config file '{filepath}' does not exist")

        # Load in the config file at the given filepath
        with open(filepath) as file_stream:
            self.config_dict = yaml.safe_load(file_stream.read())

        # Parse and validate config options
        self._parse_config_values()

    def _parse_config_values(self) -> None:
        """Read and validate each config option"""
        # Logging setup
        formatter = logging.Formatter(
            "%(asctime)s | %(name)s [%(levelname)s] %(message)s"
        )

        log_level = self._get_cfg(["logging", "level"], default="INFO")
        logger.setLevel(log_level)

        file_logging_enabled = self._get_cfg(
            ["logging", "file_logging", "enabled"], default=False
        )
        file_logging_filepath = self._get_cfg(
            ["logging", "file_logging", "filepath"], default="matrix-alertbot.log"
        )
        if file_logging_enabled:
            file_handler = logging.FileHandler(file_logging_filepath)
            file_handler.setFormatter(formatter)
            logger.addHandler(file_handler)

        console_logging_enabled = self._get_cfg(
            ["logging", "console_logging", "enabled"], default=True
        )
        if console_logging_enabled:
            console_handler = logging.StreamHandler(sys.stdout)
            console_handler.setFormatter(formatter)
            logger.addHandler(console_handler)

        # Storage setup
        self.store_dir: str = self._get_cfg(["storage", "path"], required=True)

        # Create the store folder if it doesn't exist
        if not os.path.isdir(self.store_dir):
            if not os.path.exists(self.store_dir):
                os.mkdir(self.store_dir)
            else:
                raise InvalidConfigError(
                    f"storage.path '{self.store_dir}' is not a directory"
                )

        # Template setup
        self.template_dir: str = self._get_cfg(["template", "path"], required=False)

        # Cache setup
        self.cache_dir: str = self._get_cfg(["cache", "path"], required=True)
        expire_time: str = self._get_cfg(["cache", "expire_time"], default="1w")
        self.cache_expire_time = pytimeparse2.parse(expire_time)

        # Alertmanager client setup
        self.alertmanager_url: str = self._get_cfg(
            ["alertmanager", "url"], required=True
        )

        # Matrix bot account setup
        self.user_id: str = self._get_cfg(["matrix", "user_id"], required=True)
        if not re.match("@.+:.+", self.user_id):
            raise InvalidConfigError("matrix.user_id must be in the form @name:domain")

        self.user_password: str = self._get_cfg(
            ["matrix", "user_password"], required=False
        )
        self.user_token: str = self._get_cfg(["matrix", "user_token"], required=False)
        if not self.user_token and not self.user_password:
            raise RequiredConfigKeyError("Must supply either user token or password")

        self.device_id: str = self._get_cfg(["matrix", "device_id"], required=False)
        self.device_name: str = self._get_cfg(
            ["matrix", "device_name"], default="matrix-alertbot"
        )
        self.user_token_file: str = self._get_cfg(
            ["matrix", "user_token_file"], default="token.json"
        )
        self.homeserver_url: str = self._get_cfg(["matrix", "url"], required=True)
        self.allowed_rooms: list = self._get_cfg(
            ["matrix", "allowed_rooms"], required=True
        )
        self.allowed_reactions = set(
            self._get_cfg(["matrix", "allowed_reactions"], default=DEFAULT_REACTIONS)
        )

        self.address: str = self._get_cfg(["webhook", "address"], required=False)
        self.port: int = self._get_cfg(["webhook", "port"], required=False)
        self.socket: str = self._get_cfg(["webhook", "socket"], required=False)
        if (
            not (self.address or self.port or self.socket)
            or (self.address and not self.port)
            or (not self.address and self.port)
        ):
            raise RequiredConfigKeyError(
                "Must supply either webhook.socket or both webhook.address and webhook.port"
            )
        elif self.socket and self.address and self.port:
            raise InvalidConfigError(
                "Supplied both webhook.socket and both webhook.address"
            )

        self.command_prefix: str = (
            self._get_cfg(["command_prefix"], default="!alert") + " "
        )

    def _get_cfg(
        self,
        path: List[str],
        default: Optional[Any] = None,
        required: bool = True,
    ) -> Any:
        """Get a config option from a path and option name, specifying whether it is
        required.

        Raises:
            RequiredConfigKeyError: If required is True and the object is not found (and there is
                no default value provided), a ConfigError will be raised.
        """
        # Sift through the the config until we reach our option
        config = self.config_dict
        for name in path:
            config = config.get(name)

            # If at any point we don't get our expected option...
            if config is None:
                # Raise an error if it was required
                if required and default is None:
                    raise RequiredConfigKeyError(
                        f"Config option {'.'.join(path)} is required"
                    )

                # or return the default value
                return default

        # We found the option. Return it.
        return config



import logging
from typing import Dict, Optional, TypedDict, Union

from nio import (
    AsyncClient,
    ErrorResponse,
    Response,
    RoomSendError,
    RoomSendResponse,
    SendRetryError,
)
from typing_extensions import NotRequired

logger = logging.getLogger(__name__)


ContentEventDict = TypedDict(
    "ContentEventDict",
    {
        "msgtype": str,
        "format": str,
        "body": str,
        "formatted_body": NotRequired[str],
        "m.relates_to": NotRequired[Dict],
    },
)


async def send_text_to_room(
    matrix_client: AsyncClient,
    room_id: str,
    plaintext: str,
    html: str = None,
    notice: bool = True,
    reply_to_event_id: Optional[str] = None,
) -> RoomSendResponse:
    """Send text to a matrix room.

    Args:
        client: The client to communicate to matrix with.

        room_id: The ID of the room to send the message to.

        plaintext: The message content.

        html: The message content in HTML format.

        notice: Whether the message should be sent with an "m.notice" message type
            (will not ping users).

        reply_to_event_id: Whether this message is a reply to another event. The event
            ID this is message is a reply to.

    Returns:
        A RoomSendResponse if the request was successful, else an ErrorResponse.
    """
    # Determine whether to ping room members or not
    msgtype = "m.notice" if notice else "m.text"

    content: ContentEventDict = {
        "msgtype": msgtype,
        "format": "org.matrix.custom.html",
        "body": plaintext,
    }

    if html is not None:
        content["formatted_body"] = html

    if reply_to_event_id:
        content["m.relates_to"] = {"m.in_reply_to": {"event_id": reply_to_event_id}}

    response_event = await matrix_client.room_send(
        room_id,
        "m.room.message",
        content,
        ignore_unverified_devices=True,
    )

    if isinstance(response_event, RoomSendError):
        raise SendRetryError(response_event.error)
    return response_event


async def react_to_event(
    client: AsyncClient,
    room_id: str,
    event_id: str,
    reaction_text: str,
) -> Union[Response, ErrorResponse]:
    """Reacts to a given event in a room with the given reaction text

    Args:
        client: The client to communicate to matrix with.

        room_id: The ID of the room to send the message to.

        event_id: The ID of the event to react to.

        reaction_text: The string to react with. Can also be (one or more) emoji characters.

    Returns:
        A nio.Response or nio.ErrorResponse if an error occurred.

    Raises:
        SendRetryError: If the reaction was unable to be sent.
    """
    content = {
        "m.relates_to": {
            "rel_type": "m.annotation",
            "event_id": event_id,
            "key": reaction_text,
        }
    }

    response_event = await client.room_send(
        room_id,
        "m.reaction",
        content,
        ignore_unverified_devices=True,
    )

    if isinstance(response_event, RoomSendError):
        raise SendRetryError(response_event.error)
    return response_event


def strip_fallback(content: str) -> str:
    index = 0
    for line in content.splitlines(keepends=True):
        if not line.startswith("> "):
            break
        if index == 0:
            index += 1
        index += len(line)
    return content[index:]



import unittest
from typing import Any, Dict, Optional
from unittest.mock import Mock

import nio

from matrix_alertbot.chat_functions import (
    react_to_event,
    send_text_to_room,
    strip_fallback,
)

from tests.utils import make_awaitable


async def send_room_raise_send_retry_error(
    room_id: str,
    message_type: str,
    content: Dict[Any, Any],
    tx_id: Optional[str] = None,
    ignore_unverified_devices: bool = False,
) -> nio.RoomSendResponse:
    raise nio.SendRetryError


class ChatFunctionsTestCase(unittest.IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        pass

    def test_strip_fallback(self) -> None:
        fake_body = "> some message\n\nsome reply"
        message = strip_fallback(fake_body)
        self.assertEqual("some reply", message)

        fake_body = "some message"
        message = strip_fallback(fake_body)
        self.assertEqual(fake_body, message)

    async def test_react_to_event(self) -> None:
        fake_response = Mock(spec=nio.RoomSendResponse)
        fake_matrix_client = Mock(spec=nio.AsyncClient)
        fake_matrix_client.room_send = Mock(return_value=make_awaitable(fake_response))
        fake_room_id = "!abcdefgh:example.com"
        fake_event_id = "some event id"
        fake_reaction_text = "some reaction"

        response = await react_to_event(
            fake_matrix_client, fake_room_id, fake_event_id, fake_reaction_text
        )

        fake_matrix_client.room_send.assert_called_once_with(
            fake_room_id,
            "m.reaction",
            {
                "m.relates_to": {
                    "rel_type": "m.annotation",
                    "event_id": fake_event_id,
                    "key": fake_reaction_text,
                }
            },
            ignore_unverified_devices=True,
        )
        self.assertEqual(fake_response, response)

    async def test_react_to_event_return_room_send_error(self) -> None:
        fake_response = Mock(spec=nio.RoomSendError)
        fake_response.error = "some error"
        fake_matrix_client = Mock(spec=nio.AsyncClient)
        fake_matrix_client.room_send.return_value = make_awaitable(fake_response)
        fake_room_id = "!abcdefgh:example.com"
        fake_event_id = "some event id"
        fake_reaction_text = "some reaction"

        with self.assertRaises(nio.SendRetryError):
            await react_to_event(
                fake_matrix_client, fake_room_id, fake_event_id, fake_reaction_text
            )

        fake_matrix_client.room_send.assert_called_once_with(
            fake_room_id,
            "m.reaction",
            {
                "m.relates_to": {
                    "rel_type": "m.annotation",
                    "event_id": fake_event_id,
                    "key": fake_reaction_text,
                }
            },
            ignore_unverified_devices=True,
        )

    async def test_send_text_to_room_as_notice(self) -> None:
        fake_response = Mock(spec=nio.RoomSendResponse)
        fake_matrix_client = Mock(spec=nio.AsyncClient)
        fake_matrix_client.room_send = Mock(return_value=make_awaitable(fake_response))
        fake_room_id = "!abcdefgh:example.com"
        fake_plaintext_body = "some plaintext message"
        fake_html_body = "some html message"

        response = await send_text_to_room(
            fake_matrix_client, fake_room_id, fake_plaintext_body, fake_html_body
        )

        fake_matrix_client.room_send.assert_called_once_with(
            fake_room_id,
            "m.room.message",
            {
                "msgtype": "m.notice",
                "format": "org.matrix.custom.html",
                "body": fake_plaintext_body,
                "formatted_body": fake_html_body,
            },
            ignore_unverified_devices=True,
        )
        self.assertEqual(fake_response, response)

    async def test_send_text_to_room_as_message(self) -> None:
        fake_response = Mock(spec=nio.RoomSendResponse)
        fake_matrix_client = Mock(spec=nio.AsyncClient)
        fake_matrix_client.room_send.return_value = make_awaitable(fake_response)
        fake_room_id = "!abcdefgh:example.com"
        fake_plaintext_body = "some plaintext message"
        fake_html_body = "some html message"

        response = await send_text_to_room(
            fake_matrix_client,
            fake_room_id,
            fake_plaintext_body,
            fake_html_body,
            notice=False,
        )

        fake_matrix_client.room_send.assert_called_once_with(
            fake_room_id,
            "m.room.message",
            {
                "msgtype": "m.text",
                "format": "org.matrix.custom.html",
                "body": fake_plaintext_body,
                "formatted_body": fake_html_body,
            },
            ignore_unverified_devices=True,
        )
        self.assertEqual(fake_response, response)

    async def test_send_text_to_room_in_reply_to_event(self) -> None:
        fake_response = Mock(spec=nio.RoomSendResponse)
        fake_matrix_client = Mock(spec=nio.AsyncClient)
        fake_matrix_client.room_send.return_value = make_awaitable(fake_response)
        fake_room_id = "!abcdefgh:example.com"
        fake_plaintext_body = "some plaintext message"
        fake_html_body = "some html message"
        fake_event_id = "some event id"

        response = await send_text_to_room(
            fake_matrix_client,
            fake_room_id,
            fake_plaintext_body,
            fake_html_body,
            reply_to_event_id=fake_event_id,
        )

        fake_matrix_client.room_send.assert_called_once_with(
            fake_room_id,
            "m.room.message",
            {
                "msgtype": "m.notice",
                "format": "org.matrix.custom.html",
                "body": fake_plaintext_body,
                "formatted_body": fake_html_body,
                "m.relates_to": {"m.in_reply_to": {"event_id": fake_event_id}},
            },
            ignore_unverified_devices=True,
        )
        self.assertEqual(fake_response, response)

    async def test_send_text_to_room_raise_send_retry_error(self) -> None:
        fake_matrix_client = Mock(spec=nio.AsyncClient)
        fake_matrix_client.room_send.side_effect = send_room_raise_send_retry_error

        fake_room_id = "!abcdefgh:example.com"
        fake_plaintext_body = "some plaintext message"
        fake_html_body = "some html message"

        with self.assertRaises(nio.SendRetryError):
            await send_text_to_room(
                fake_matrix_client,
                fake_room_id,
                fake_plaintext_body,
                fake_html_body,
            )

        fake_matrix_client.room_send.assert_called_once_with(
            fake_room_id,
            "m.room.message",
            {
                "msgtype": "m.notice",
                "format": "org.matrix.custom.html",
                "body": fake_plaintext_body,
                "formatted_body": fake_html_body,
            },
            ignore_unverified_devices=True,
        )

    async def test_send_text_to_room_return_room_send_error(self) -> None:
        fake_response = Mock(spec=nio.RoomSendError)
        fake_response.error = "some error"
        fake_matrix_client = Mock(spec=nio.AsyncClient)
        fake_matrix_client.room_send.return_value = make_awaitable(fake_response)
        fake_room_id = "!abcdefgh:example.com"
        fake_plaintext_body = "some plaintext message"
        fake_html_body = "some html message"

        with self.assertRaises(nio.SendRetryError):
            await send_text_to_room(
                fake_matrix_client,
                fake_room_id,
                fake_plaintext_body,
                fake_html_body,
            )

        fake_matrix_client.room_send.assert_called_once_with(
            fake_room_id,
            "m.room.message",
            {
                "msgtype": "m.notice",
                "format": "org.matrix.custom.html",
                "body": fake_plaintext_body,
                "formatted_body": fake_html_body,
            },
            ignore_unverified_devices=True,
        )


if __name__ == "__main__":
    unittest.main()



import os
import unittest
from datetime import timedelta
from unittest.mock import Mock, patch

import yaml

from matrix_alertbot.config import DEFAULT_REACTIONS, Config
from matrix_alertbot.errors import (
    InvalidConfigError,
    ParseConfigError,
    RequiredConfigKeyError,
)

WORKING_DIR = os.path.dirname(__file__)
CONFIG_RESOURCES_DIR = os.path.join(WORKING_DIR, "resources", "config")


class DummyConfig(Config):
    def __init__(self, filepath: str):
        with open(filepath) as file_stream:
            self.config_dict = yaml.safe_load(file_stream.read())


def mock_path_isdir(path: str) -> bool:
    if path == "data/store":
        return False
    return True


def mock_path_exists(path: str) -> bool:
    if path == "data/store":
        return False
    return True


class ConfigTestCase(unittest.TestCase):
    @patch("os.path.isdir")
    @patch("os.path.exists")
    @patch("os.mkdir")
    def test_read_minimal_config(
        self, fake_mkdir: Mock, fake_path_exists: Mock, fake_path_isdir: Mock
    ) -> None:
        fake_path_isdir.return_value = False
        fake_path_exists.return_value = False

        config_path = os.path.join(CONFIG_RESOURCES_DIR, "config.minimal.yml")
        config = Config(config_path)

        fake_path_isdir.assert_called_once_with("data/store")
        fake_path_exists.assert_called_once_with("data/store")
        fake_mkdir.assert_called_once_with("data/store")

        self.assertEqual("@fakes_user:matrix.example.com", config.user_id)
        self.assertEqual("password", config.user_password)
        self.assertIsNone(config.user_token)
        self.assertIsNone(config.device_id)
        self.assertEqual("matrix-alertbot", config.device_name)
        self.assertEqual("https://matrix.example.com", config.homeserver_url)
        self.assertEqual(["!abcdefgh:matrix.example.com"], config.allowed_rooms)
        self.assertEqual(DEFAULT_REACTIONS, config.allowed_reactions)

        self.assertEqual("0.0.0.0", config.address)
        self.assertEqual(8080, config.port)
        self.assertIsNone(config.socket)

        self.assertEqual("http://localhost:9093", config.alertmanager_url)

        expected_expire_time = timedelta(days=7).total_seconds()
        self.assertEqual(expected_expire_time, config.cache_expire_time)
        self.assertEqual("data/cache", config.cache_dir)

        self.assertEqual("data/store", config.store_dir)

        self.assertIsNone(config.template_dir)

        self.assertEqual("!alert ", config.command_prefix)

    @patch("os.path.isdir")
    @patch("os.path.exists")
    @patch("os.mkdir")
    def test_read_full_config(
        self, fake_mkdir: Mock, fake_path_exists: Mock, fake_path_isdir: Mock
    ) -> None:
        fake_path_isdir.return_value = False
        fake_path_exists.return_value = False

        config_path = os.path.join(CONFIG_RESOURCES_DIR, "config.full.yml")
        config = Config(config_path)

        fake_path_isdir.assert_called_once_with("data/store")
        fake_path_exists.assert_called_once_with("data/store")
        fake_mkdir.assert_called_once_with("data/store")

        self.assertEqual("@fakes_user:matrix.example.com", config.user_id)
        self.assertEqual("password", config.user_password)
        self.assertIsNone(config.user_token)
        self.assertEqual("token.json", config.user_token_file)
        self.assertEqual("ABCDEFGHIJ", config.device_id)
        self.assertEqual("fake_device_name", config.device_name)
        self.assertEqual("https://matrix.example.com", config.homeserver_url)
        self.assertEqual(["!abcdefgh:matrix.example.com"], config.allowed_rooms)
        self.assertEqual({"🤫", "😶", "🤐"}, config.allowed_reactions)

        self.assertIsNone(config.address)
        self.assertIsNone(config.port)
        self.assertEqual("matrix-alertbot.socket", config.socket)

        self.assertEqual("http://localhost:9093", config.alertmanager_url)

        expected_expire_time = timedelta(days=7).total_seconds()
        self.assertEqual(expected_expire_time, config.cache_expire_time)
        self.assertEqual("data/cache", config.cache_dir)

        self.assertEqual("data/store", config.store_dir)

        self.assertEqual("data/templates", config.template_dir)

        self.assertEqual("!alert ", config.command_prefix)

    def test_read_config_raise_config_error(self) -> None:
        with self.assertRaises(ParseConfigError):
            Config("")

    @patch("os.path.isdir")
    @patch("os.path.exists")
    @patch("os.mkdir")
    def test_parse_config_with_storage_path_error(
        self, fake_mkdir: Mock, fake_path_exists: Mock, fake_path_isdir: Mock
    ) -> None:
        fake_path_isdir.return_value = False
        fake_path_exists.return_value = True

        config_path = os.path.join(CONFIG_RESOURCES_DIR, "config.minimal.yml")
        with self.assertRaises(ParseConfigError):
            Config(config_path)

        fake_path_isdir.assert_called_once_with("data/store")
        fake_path_exists.assert_called_once_with("data/store")
        fake_mkdir.assert_not_called()

    @patch("os.path.isdir")
    @patch("os.path.exists")
    @patch("os.mkdir")
    def test_parse_config_with_missing_matrix_user_id(
        self, fake_mkdir: Mock, fake_path_exists: Mock, fake_path_isdir: Mock
    ) -> None:
        fake_path_isdir.return_value = False
        fake_path_exists.return_value = False

        config_path = os.path.join(CONFIG_RESOURCES_DIR, "config.minimal.yml")
        config = DummyConfig(config_path)
        del config.config_dict["matrix"]["user_id"]

        with self.assertRaises(RequiredConfigKeyError):
            config._parse_config_values()

    @patch("os.path.isdir")
    @patch("os.path.exists")
    @patch("os.mkdir")
    def test_parse_config_with_missing_matrix_user_password(
        self, fake_mkdir: Mock, fake_path_exists: Mock, fake_path_isdir: Mock
    ) -> None:
        fake_path_isdir.return_value = False
        fake_path_exists.return_value = False

        config_path = os.path.join(CONFIG_RESOURCES_DIR, "config.minimal.yml")
        config = DummyConfig(config_path)
        del config.config_dict["matrix"]["user_password"]

        with self.assertRaises(RequiredConfigKeyError):
            config._parse_config_values()

    @patch("os.path.isdir")
    @patch("os.path.exists")
    @patch("os.mkdir")
    def test_parse_config_with_missing_matrix_url(
        self, fake_mkdir: Mock, fake_path_exists: Mock, fake_path_isdir: Mock
    ) -> None:
        fake_path_isdir.return_value = False
        fake_path_exists.return_value = False

        config_path = os.path.join(CONFIG_RESOURCES_DIR, "config.minimal.yml")
        config = DummyConfig(config_path)
        del config.config_dict["matrix"]["url"]

        with self.assertRaises(RequiredConfigKeyError):
            config._parse_config_values()

    @patch("os.path.isdir")
    @patch("os.path.exists")
    @patch("os.mkdir")
    def test_parse_config_with_missing_matrix_allowed_rooms(
        self, fake_mkdir: Mock, fake_path_exists: Mock, fake_path_isdir: Mock
    ) -> None:
        fake_path_isdir.return_value = False
        fake_path_exists.return_value = False

        config_path = os.path.join(CONFIG_RESOURCES_DIR, "config.minimal.yml")
        config = DummyConfig(config_path)
        del config.config_dict["matrix"]["allowed_rooms"]

        with self.assertRaises(RequiredConfigKeyError):
            config._parse_config_values()

    @patch("os.path.isdir")
    @patch("os.path.exists")
    @patch("os.mkdir")
    def test_parse_config_with_missing_webhook_address(
        self, fake_mkdir: Mock, fake_path_exists: Mock, fake_path_isdir: Mock
    ) -> None:
        fake_path_isdir.return_value = False
        fake_path_exists.return_value = False

        config_path = os.path.join(CONFIG_RESOURCES_DIR, "config.minimal.yml")
        config = DummyConfig(config_path)
        del config.config_dict["webhook"]["address"]

        with self.assertRaises(RequiredConfigKeyError):
            config._parse_config_values()

    @patch("os.path.isdir")
    @patch("os.path.exists")
    @patch("os.mkdir")
    def test_parse_config_with_missing_alertmanager_url(
        self, fake_mkdir: Mock, fake_path_exists: Mock, fake_path_isdir: Mock
    ) -> None:
        fake_path_isdir.return_value = False
        fake_path_exists.return_value = False

        config_path = os.path.join(CONFIG_RESOURCES_DIR, "config.minimal.yml")
        config = DummyConfig(config_path)
        del config.config_dict["alertmanager"]["url"]

        with self.assertRaises(RequiredConfigKeyError):
            config._parse_config_values()

    @patch("os.path.isdir")
    @patch("os.path.exists")
    @patch("os.mkdir")
    def test_parse_config_with_missing_cache_path(
        self, fake_mkdir: Mock, fake_path_exists: Mock, fake_path_isdir: Mock
    ) -> None:
        fake_path_isdir.return_value = False
        fake_path_exists.return_value = False

        config_path = os.path.join(CONFIG_RESOURCES_DIR, "config.minimal.yml")
        config = DummyConfig(config_path)
        del config.config_dict["cache"]["path"]

        with self.assertRaises(RequiredConfigKeyError):
            config._parse_config_values()

    @patch("os.path.isdir")
    @patch("os.path.exists")
    @patch("os.mkdir")
    def test_parse_config_with_missing_storage_path(
        self, fake_mkdir: Mock, fake_path_exists: Mock, fake_path_isdir: Mock
    ) -> None:
        fake_path_isdir.return_value = False
        fake_path_exists.return_value = False

        config_path = os.path.join(CONFIG_RESOURCES_DIR, "config.minimal.yml")
        config = DummyConfig(config_path)
        del config.config_dict["storage"]["path"]

        with self.assertRaises(RequiredConfigKeyError):
            config._parse_config_values()

    @patch("os.path.isdir")
    @patch("os.path.exists")
    @patch("os.mkdir")
    def test_parse_config_with_invalid_matrix_user_id(
        self, fake_mkdir: Mock, fake_path_exists: Mock, fake_path_isdir: Mock
    ) -> None:
        fake_path_isdir.return_value = False
        fake_path_exists.return_value = False

        config_path = os.path.join(CONFIG_RESOURCES_DIR, "config.minimal.yml")
        config = DummyConfig(config_path)

        config.config_dict["matrix"]["user_id"] = ""
        with self.assertRaises(InvalidConfigError):
            config._parse_config_values()

        config.config_dict["matrix"]["user_id"] = "@fake_user"
        with self.assertRaises(InvalidConfigError):
            config._parse_config_values()

        config.config_dict["matrix"]["user_id"] = "@fake_user:"
        with self.assertRaises(InvalidConfigError):
            config._parse_config_values()

        config.config_dict["matrix"]["user_id"] = ":matrix.example.com"
        with self.assertRaises(InvalidConfigError):
            config._parse_config_values()

        config.config_dict["matrix"]["user_id"] = "@:matrix.example.com"
        with self.assertRaises(InvalidConfigError):
            config._parse_config_values()

        config.config_dict["matrix"]["user_id"] = "@:"
        with self.assertRaises(InvalidConfigError):
            config._parse_config_values()

    @patch("os.path.isdir")
    @patch("os.path.exists")
    @patch("os.mkdir")
    def test_parse_config_with_both_webhook_socket_and_address(
        self, fake_mkdir: Mock, fake_path_exists: Mock, fake_path_isdir: Mock
    ) -> None:
        fake_path_isdir.return_value = False
        fake_path_exists.return_value = False

        config_path = os.path.join(CONFIG_RESOURCES_DIR, "config.minimal.yml")
        config = DummyConfig(config_path)
        config.config_dict["webhook"]["socket"] = "matrix-alertbot.socket"

        with self.assertRaises(InvalidConfigError):
            config._parse_config_values()


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import logging

import prometheus_client
from aiohttp import ClientError, web, web_request
from aiohttp_prometheus_exporter.handler import metrics
from aiohttp_prometheus_exporter.middleware import prometheus_middleware_factory
from diskcache import Cache
from nio import AsyncClient, LocalProtocolError, SendRetryError

from matrix_alertbot.alert import Alert, AlertRenderer
from matrix_alertbot.alertmanager import AlertmanagerClient
from matrix_alertbot.chat_functions import send_text_to_room
from matrix_alertbot.config import Config
from matrix_alertbot.errors import (
    AlertmanagerError,
    SilenceExtendError,
    SilenceNotFoundError,
)

logger = logging.getLogger(__name__)

routes = web.RouteTableDef()


@routes.get("/health")
async def get_health(request: web_request.Request) -> web.Response:
    return web.Response(status=200)


@routes.post("/alerts/{room_id}")
async def create_alerts(request: web_request.Request) -> web.Response:
    data = await request.json()
    room_id = request.match_info["room_id"]

    config: Config = request.app["config"]

    if room_id not in config.allowed_rooms:
        logger.error("Cannot send alerts to room ID {room_id}.")
        return web.Response(
            status=401, body=f"Cannot send alerts to room ID {room_id}."
        )

    if "alerts" not in data:
        logger.error("Received data without 'alerts' key")
        return web.Response(status=400, body="Data must contain 'alerts' key.")

    alert_dicts = data["alerts"]

    if not isinstance(data["alerts"], list):
        alerts_type = type(alert_dicts).__name__
        logger.error(f"Received data with invalid alerts type '{alerts_type}'.")
        return web.Response(
            status=400, body=f"Alerts must be a list, got '{alerts_type}'."
        )

    logger.info(f"Received {len(alert_dicts)} alerts for room ID {room_id}: {data}")

    if len(data["alerts"]) == 0:
        return web.Response(status=400, body="Alerts cannot be empty.")

    alerts = []
    for alert in alert_dicts:
        try:
            alert = Alert.from_dict(alert)
        except KeyError as e:
            logger.error(f"Cannot parse alert dict: {e}")
            return web.Response(status=400, body=f"Invalid alert: {alert}.")
        alerts.append(alert)

    for alert in alerts:
        try:
            await create_alert(alert, room_id, request)
        except AlertmanagerError as e:
            logger.error(
                f"An error occured with Alertmanager when handling alert with fingerprint {alert.fingerprint}: {e}"
            )
            return web.Response(
                status=500,
                body=f"An error occured with Alertmanager when handling alert with fingerprint {alert.fingerprint}.",
            )
        except (SendRetryError, LocalProtocolError, ClientError) as e:
            logger.error(
                f"Unable to send alert {alert.fingerprint} to Matrix room: {e}"
            )
            return web.Response(
                status=500,
                body=f"An error occured when sending alert with fingerprint '{alert.fingerprint}' to Matrix room.",
            )

    return web.Response(status=200)


async def create_alert(
    alert: Alert, room_id: str, request: web_request.Request
) -> None:
    alertmanager_client: AlertmanagerClient = request.app["alertmanager_client"]
    alert_renderer: AlertRenderer = request.app["alert_renderer"]
    matrix_client: AsyncClient = request.app["matrix_client"]
    cache: Cache = request.app["cache"]
    config: Config = request.app["config"]

    if alert.firing:
        try:
            silence_id = await alertmanager_client.update_silence(alert.fingerprint)
            logger.debug(
                f"Extended silence ID {silence_id} for alert with fingerprint {alert.fingerprint}"
            )
            return
        except SilenceNotFoundError as e:
            logger.debug(
                f"Unable to extend silence for alert with fingerprint {alert.fingerprint}: {e}"
            )
            cache.delete(alert.fingerprint)
        except SilenceExtendError as e:
            logger.debug(
                f"Unable to extend silence for alert with fingerprint {alert.fingerprint}: {e}"
            )

    plaintext = alert_renderer.render(alert, html=False)
    html = alert_renderer.render(alert, html=True)

    event = await send_text_to_room(
        matrix_client, room_id, plaintext, html, notice=False
    )

    if alert.firing:
        cache.set(event.event_id, alert.fingerprint, expire=config.cache_expire_time)
    else:
        cache.delete(alert.fingerprint)


class Webhook:
    def __init__(
        self,
        matrix_client: AsyncClient,
        alertmanager_client: AlertmanagerClient,
        cache: Cache,
        config: Config,
    ) -> None:
        self.app = web.Application(logger=logger)
        self.app["matrix_client"] = matrix_client
        self.app["alertmanager_client"] = alertmanager_client
        self.app["config"] = config
        self.app["cache"] = cache
        self.app["alert_renderer"] = AlertRenderer(config.template_dir)
        self.app.add_routes(routes)

        prometheus_registry = prometheus_client.CollectorRegistry(auto_describe=True)
        self.app.middlewares.append(
            prometheus_middleware_factory(registry=prometheus_registry)
        )
        self.app.router.add_get("/metrics", metrics())

        self.runner = web.AppRunner(self.app)

        self.config = config
        self.address = config.address
        self.port = config.port
        self.socket = config.socket

    async def start(self) -> None:
        await self.runner.setup()

        site: web.BaseSite
        if self.address and self.port:
            site = web.TCPSite(self.runner, self.address, self.port)
            logger.info(f"Listening on {self.address}:{self.port}")
        elif self.socket:
            site = web.UnixSite(self.runner, self.socket)
            logger.info(f"Listening on unix://{self.socket}")

        await site.start()

    async def close(self) -> None:
        await self.runner.cleanup()



from __future__ import annotations

import json
import unittest
from datetime import datetime, timedelta
from typing import Any, Dict, Optional, Tuple
from unittest.mock import Mock, patch

import aiohttp
import aiohttp.test_utils
import aiotools
from aiohttp import web, web_request
from diskcache import Cache
from freezegun import freeze_time

import matrix_alertbot.alertmanager
from matrix_alertbot.alertmanager import AlertmanagerClient
from matrix_alertbot.errors import (
    AlertmanagerServerError,
    AlertNotFoundError,
    SilenceExpiredError,
    SilenceExtendError,
    SilenceNotFoundError,
)


async def update_silence_raise_silence_not_found(
    fingerprint: str, user: str, duration_seconds: int, *, force: bool = False
) -> str:
    raise SilenceNotFoundError


class FakeCache:
    def __init__(self, cache_dict: Optional[Dict] = None) -> None:
        if cache_dict is None:
            cache_dict = {}
        self.cache = cache_dict

    def get(
        self, key: str, expire_time: bool = False
    ) -> Optional[Tuple[str, Optional[int]] | str]:
        return self.cache.get(key)

    def set(self, key: str, value: str, expire: int) -> None:
        self.cache[key] = value, expire


class AbstractFakeAlertmanagerServer:
    def __init__(self) -> None:
        self.app = web.Application()
        self.app.router.add_routes(
            [
                web.get("/api/v2/alerts", self.get_alerts),
                web.get("/api/v2/silences", self.get_silences),
                web.post("/api/v2/silences", self.create_silence),
                web.delete("/api/v2/silence/{silence}", self.delete_silence),
            ]
        )
        self.app["silences"] = [
            {
                "id": "silence1",
                "createdBy": "user1",
                "status": {"state": "active"},
                "matchers": [],
            },
            {
                "id": "silence2",
                "createdBy": "user2",
                "status": {"state": "expired"},
                "matchers": [],
            },
        ]

        self.runner = web.AppRunner(self.app)

    async def __aenter__(self) -> AbstractFakeAlertmanagerServer:
        await self.start()
        return self

    async def __aexit__(self, *args: Any) -> None:
        await self.stop()

    async def start(self) -> None:
        self.port = aiohttp.test_utils.unused_port()

        await self.runner.setup()

        site = web.TCPSite(self.runner, "127.0.0.1", self.port)
        await site.start()

    async def stop(self) -> None:
        await self.runner.cleanup()

    async def get_alerts(self, request: web_request.Request) -> web.Response:
        raise NotImplementedError

    async def get_silences(self, request: web_request.Request) -> web.Response:
        raise NotImplementedError

    async def create_silence(self, request: web_request.Request) -> web.Response:
        raise NotImplementedError

    async def delete_silence(self, request: web_request.Request) -> web.Response:
        raise NotImplementedError


class FakeAlertmanagerServer(AbstractFakeAlertmanagerServer):
    async def get_alerts(self, request: web_request.Request) -> web.Response:
        return web.Response(
            body=json.dumps(
                [
                    {
                        "fingerprint": "fingerprint1",
                        "labels": {"alertname": "alert1"},
                        "status": {"state": "active"},
                    },
                    {
                        "fingerprint": "fingerprint2",
                        "labels": {"alertname": "alert2"},
                        "status": {
                            "state": "suppressed",
                            "silencedBy": ["silence1", "silence2"],
                        },
                    },
                ]
            ),
            content_type="application/json",
        )

    async def get_silences(self, request: web_request.Request) -> web.Response:
        return web.Response(
            body=json.dumps(self.app["silences"]), content_type="application/json"
        )

    async def create_silence(self, request: web_request.Request) -> web.Response:
        silences = self.app["silences"]

        silence = await request.json()

        silence_count = len(silences) + 1
        silence["id"] = f"silence{silence_count}"
        silence["status"] = {"state": "active"}
        silences.append(silence)

        return web.Response(
            body=json.dumps({"silenceID": silence["id"]}),
            content_type="application/json",
        )

    async def delete_silence(self, request: web_request.Request) -> web.Response:
        silence_id = request.match_info["silence"]
        for i, silence in enumerate(self.app["silences"]):
            if silence["id"] == silence_id:
                del self.app["silences"][i]
                break

        return web.Response(status=200, content_type="application/json")


class FakeAlertmanagerServerWithoutAlert(FakeAlertmanagerServer):
    async def get_alerts(self, request: web_request.Request) -> web.Response:
        return web.Response(body=json.dumps([]), content_type="application/json")


class FakeAlertmanagerServerWithErrorAlerts(FakeAlertmanagerServer):
    async def get_alerts(self, request: web_request.Request) -> web.Response:
        return web.Response(status=500)


class FakeAlertmanagerServerWithoutSilence(FakeAlertmanagerServer):
    def __init__(self) -> None:
        super().__init__()
        self.app["silences"] = []


class FakeAlertmanagerServerWithErrorSilences(FakeAlertmanagerServer):
    async def get_silences(self, request: web_request.Request) -> web.Response:
        return web.Response(status=500)


class FakeAlertmanagerServerWithErrorCreateSilence(FakeAlertmanagerServer):
    async def create_silence(self, request: web_request.Request) -> web.Response:
        return web.Response(status=500)


class FakeAlertmanagerServerWithErrorDeleteSilence(FakeAlertmanagerServer):
    async def delete_silence(self, request: web_request.Request) -> web.Response:
        return web.Response(status=500)


class AlertmanagerClientTestCase(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        self.fake_fingerprints = Mock(return_value=["fingerprint1", "fingerprint2"])

    async def test_get_alerts_happy(self) -> None:
        fake_cache = FakeCache()

        async with FakeAlertmanagerServer() as fake_alertmanager_server:
            port = fake_alertmanager_server.port
            alertmanager_client = AlertmanagerClient(
                f"http://localhost:{port}", fake_cache
            )
            async with aiotools.closing_async(alertmanager_client):
                alerts = await alertmanager_client.get_alerts()

        self.assertEqual(
            [
                {
                    "fingerprint": "fingerprint1",
                    "labels": {"alertname": "alert1"},
                    "status": {"state": "active"},
                },
                {
                    "fingerprint": "fingerprint2",
                    "labels": {"alertname": "alert2"},
                    "status": {
                        "state": "suppressed",
                        "silencedBy": ["silence1", "silence2"],
                    },
                },
            ],
            alerts,
        )

    async def test_get_alerts_empty(self) -> None:
        fake_cache = FakeCache()

        async with FakeAlertmanagerServerWithoutAlert() as fake_alertmanager_server:
            port = fake_alertmanager_server.port
            alertmanager_client = AlertmanagerClient(
                f"http://localhost:{port}", fake_cache
            )
            async with aiotools.closing_async(alertmanager_client):
                alerts = await alertmanager_client.get_alerts()

        self.assertEqual([], alerts)

    async def test_get_alerts_raise_alertmanager_error(self) -> None:
        fake_cache = FakeCache()

        async with FakeAlertmanagerServerWithErrorAlerts() as fake_alertmanager_server:
            port = fake_alertmanager_server.port
            alertmanager_client = AlertmanagerClient(
                f"http://localhost:{port}", fake_cache
            )
            async with aiotools.closing_async(alertmanager_client):
                with self.assertRaises(AlertmanagerServerError):
                    await alertmanager_client.get_alerts()

    async def test_get_silences_happy(self) -> None:
        fake_cache = FakeCache()

        async with FakeAlertmanagerServer() as fake_alertmanager_server:
            port = fake_alertmanager_server.port
            alertmanager_client = AlertmanagerClient(
                f"http://localhost:{port}", fake_cache
            )
            async with aiotools.closing_async(alertmanager_client):
                silences = await alertmanager_client.get_silences()

        self.assertEqual(
            [
                {
                    "id": "silence1",
                    "createdBy": "user1",
                    "status": {"state": "active"},
                    "matchers": [],
                },
                {
                    "id": "silence2",
                    "createdBy": "user2",
                    "status": {"state": "expired"},
                    "matchers": [],
                },
            ],
            silences,
        )

    async def test_get_silences_empty(self) -> None:
        fake_cache = FakeCache()

        async with FakeAlertmanagerServerWithoutSilence() as fake_alertmanager_server:
            port = fake_alertmanager_server.port
            alertmanager_client = AlertmanagerClient(
                f"http://localhost:{port}", fake_cache
            )
            async with aiotools.closing_async(alertmanager_client):
                silences = await alertmanager_client.get_silences()

        self.assertEqual([], silences)

    async def test_get_silences_raise_alertmanager_error(self) -> None:
        fake_cache = FakeCache()

        async with FakeAlertmanagerServerWithErrorSilences() as fake_alertmanager_server:
            port = fake_alertmanager_server.port
            alertmanager_client = AlertmanagerClient(
                f"http://localhost:{port}", fake_cache
            )
            async with aiotools.closing_async(alertmanager_client):
                with self.assertRaises(AlertmanagerServerError):
                    await alertmanager_client.get_silences()

    async def test_get_alert_happy(self) -> None:
        fake_cache = FakeCache()

        async with FakeAlertmanagerServer() as fake_alertmanager_server:
            port = fake_alertmanager_server.port
            alertmanager_client = AlertmanagerClient(
                f"http://localhost:{port}", fake_cache
            )
            async with aiotools.closing_async(alertmanager_client):
                alert = await alertmanager_client.get_alert("fingerprint1")

        self.assertEqual(
            {
                "fingerprint": "fingerprint1",
                "labels": {"alertname": "alert1"},
                "status": {"state": "active"},
            },
            alert,
        )

    async def test_get_alert_raise_alert_not_found(self) -> None:
        fake_cache = FakeCache()

        async with FakeAlertmanagerServerWithoutAlert() as fake_alertmanager_server:
            port = fake_alertmanager_server.port
            alertmanager_client = AlertmanagerClient(
                f"http://localhost:{port}", fake_cache
            )
            async with aiotools.closing_async(alertmanager_client):
                with self.assertRaises(AlertNotFoundError):
                    await alertmanager_client.get_alert("fingerprint1")

    async def test_get_alert_raise_alertmanager_error(self) -> None:
        fake_cache = FakeCache()

        async with FakeAlertmanagerServerWithErrorAlerts() as fake_alertmanager_server:
            port = fake_alertmanager_server.port
            alertmanager_client = AlertmanagerClient(
                f"http://localhost:{port}", fake_cache
            )
            async with aiotools.closing_async(alertmanager_client):
                with self.assertRaises(AlertmanagerServerError):
                    await alertmanager_client.get_alert("fingerprint1")

    async def test_get_silence_happy(self) -> None:
        fake_cache = FakeCache()

        async with FakeAlertmanagerServer() as fake_alertmanager_server:
            port = fake_alertmanager_server.port
            alertmanager_client = AlertmanagerClient(
                f"http://localhost:{port}", fake_cache
            )
            async with aiotools.closing_async(alertmanager_client):
                silence1 = await alertmanager_client.get_silence("silence1")
                silence2 = await alertmanager_client.get_silence("silence2")

        self.assertEqual(
            {
                "id": "silence1",
                "createdBy": "user1",
                "status": {"state": "active"},
                "matchers": [],
            },
            silence1,
        )
        self.assertEqual(
            {
                "id": "silence2",
                "createdBy": "user2",
                "status": {"state": "expired"},
                "matchers": [],
            },
            silence2,
        )

    async def test_get_silence_raise_silence_not_found(self) -> None:
        fake_cache = FakeCache()

        async with FakeAlertmanagerServerWithoutSilence() as fake_alertmanager_server:
            port = fake_alertmanager_server.port
            alertmanager_client = AlertmanagerClient(
                f"http://localhost:{port}", fake_cache
            )
            async with aiotools.closing_async(alertmanager_client):
                with self.assertRaises(SilenceNotFoundError):
                    await alertmanager_client.get_silence("silence1")

    async def test_get_silence_raise_alertmanager_error(self) -> None:
        fake_cache = FakeCache()

        async with FakeAlertmanagerServerWithErrorSilences() as fake_alertmanager_server:
            port = fake_alertmanager_server.port
            alertmanager_client = AlertmanagerClient(
                f"http://localhost:{port}", fake_cache
            )
            async with aiotools.closing_async(alertmanager_client):
                with self.assertRaises(AlertmanagerServerError):
                    await alertmanager_client.get_silence("silence1")

    @freeze_time(datetime.utcfromtimestamp(0))
    async def test_create_silence_with_duration(self) -> None:
        fake_cache = Mock(return_value=FakeCache())

        async with FakeAlertmanagerServerWithoutSilence() as fake_alertmanager_server:
            port = fake_alertmanager_server.port
            alertmanager_client = AlertmanagerClient(
                f"http://localhost:{port}", fake_cache
            )
            async with aiotools.closing_async(alertmanager_client):
                silence_id = await alertmanager_client.create_silence(
                    "fingerprint1", "user", 86400
                )
                silence = await alertmanager_client.get_silence("silence1")

        self.assertEqual("silence1", silence_id)
        self.assertEqual(
            {
                "id": "silence1",
                "status": {"state": "active"},
                "matchers": [
                    {
                        "name": "alertname",
                        "value": "alert1",
                        "isRegex": False,
                        "isEqual": True,
                    }
                ],
                "createdBy": "user",
                "startsAt": "1970-01-01T00:00:00",
                "endsAt": "1970-01-02T00:00:00",
                "comment": "Acknowledge alert from Matrix",
            },
            silence,
        )
        fake_cache.set.assert_called_once_with("fingerprint1", "silence1", expire=86400)

    @freeze_time(datetime.utcfromtimestamp(0))
    async def test_update_silence_raise_extend_error(self) -> None:
        fake_cache = FakeCache()

        async with FakeAlertmanagerServerWithoutSilence() as fake_alertmanager_server:
            port = fake_alertmanager_server.port
            alertmanager_client = AlertmanagerClient(
                f"http://localhost:{port}", fake_cache
            )
            async with aiotools.closing_async(alertmanager_client):
                await alertmanager_client.create_silence("fingerprint1", "user", 86400)
                with self.assertRaises(SilenceExtendError):
                    await alertmanager_client.update_silence("fingerprint1")
                with self.assertRaises(SilenceNotFoundError):
                    await alertmanager_client.get_silence("silence2")
        self.assertEqual({"fingerprint1": ("silence1", 86400)}, fake_cache.cache)

    @freeze_time(datetime.utcfromtimestamp(0))
    async def test_update_silence_remove_duration(self) -> None:
        fake_cache = FakeCache()

        async with FakeAlertmanagerServerWithoutSilence() as fake_alertmanager_server:
            port = fake_alertmanager_server.port
            alertmanager_client = AlertmanagerClient(
                f"http://localhost:{port}", fake_cache
            )
            async with aiotools.closing_async(alertmanager_client):
                silence_id1 = await alertmanager_client.create_silence(
                    "fingerprint1", "user", 86400
                )
                silence_id2 = await alertmanager_client.update_silence(
                    "fingerprint1", force=True
                )
                silence2 = await alertmanager_client.get_silence("silence2")

        self.assertEqual("silence1", silence_id1)
        self.assertEqual("silence2", silence_id2)
        self.assertEqual(
            {
                "id": "silence2",
                "status": {"state": "active"},
                "matchers": [
                    {
                        "name": "alertname",
                        "value": "alert1",
                        "isRegex": False,
                        "isEqual": True,
                    }
                ],
                "createdBy": "user",
                "startsAt": "1970-01-01T00:00:00",
                "endsAt": "1970-01-01T03:00:00",
                "comment": "Acknowledge alert from Matrix",
            },
            silence2,
        )
        self.assertEqual({"fingerprint1": ("silence2", None)}, fake_cache.cache)

    @freeze_time(datetime.utcfromtimestamp(0))
    async def test_update_silence_override_user_and_duration(self) -> None:
        fake_cache = FakeCache()

        async with FakeAlertmanagerServerWithoutSilence() as fake_alertmanager_server:
            port = fake_alertmanager_server.port
            alertmanager_client = AlertmanagerClient(
                f"http://localhost:{port}", fake_cache
            )
            async with aiotools.closing_async(alertmanager_client):
                await alertmanager_client.create_silence("fingerprint1", "user1", 86400)
                silence_id2 = await alertmanager_client.update_silence(
                    "fingerprint1", "user2", 864000
                )
                silence2 = await alertmanager_client.get_silence("silence2")

        self.assertEqual("silence2", silence_id2)
        self.assertEqual(
            {
                "id": "silence2",
                "status": {"state": "active"},
                "matchers": [
                    {
                        "name": "alertname",
                        "value": "alert1",
                        "isRegex": False,
                        "isEqual": True,
                    }
                ],
                "createdBy": "user2",
                "startsAt": "1970-01-01T00:00:00",
                "endsAt": "1970-01-11T00:00:00",
                "comment": "Acknowledge alert from Matrix",
            },
            silence2,
        )
        self.assertEqual({"fingerprint1": ("silence2", 864000)}, fake_cache.cache)

    @patch.object(matrix_alertbot.alertmanager.AlertmanagerClient, "update_silence")
    @patch.object(matrix_alertbot.alertmanager.AlertmanagerClient, "create_silence")
    async def test_create_or_update_silence_with_duration_and_silence_not_found(
        self, fake_create_silence: Mock, fake_update_silence: Mock
    ) -> None:
        fake_cache = Mock(spec=Cache)
        fake_update_silence.side_effect = update_silence_raise_silence_not_found
        fake_create_silence.return_value = "silence1"

        alertmanager_client = AlertmanagerClient("http://localhost", fake_cache)
        async with aiotools.closing_async(alertmanager_client):
            silence_id1 = await alertmanager_client.create_or_update_silence(
                "fingerprint1", "user", 86400
            )

        self.assertEqual("silence1", silence_id1)
        fake_update_silence.assert_called_once_with(
            "fingerprint1", "user", 86400, force=False
        )
        fake_create_silence.assert_called_once_with("fingerprint1", "user", 86400)

    @patch.object(matrix_alertbot.alertmanager.AlertmanagerClient, "update_silence")
    @patch.object(matrix_alertbot.alertmanager.AlertmanagerClient, "create_silence")
    async def test_create_or_update_silence_with_duration_and_silence_found(
        self, fake_create_silence: Mock, fake_update_silence: Mock
    ) -> None:
        fake_cache = Mock(spec=Cache)
        fake_update_silence.return_value = "silence1"

        alertmanager_client = AlertmanagerClient("http://localhost", fake_cache)
        async with aiotools.closing_async(alertmanager_client):
            silence_id1 = await alertmanager_client.create_or_update_silence(
                "fingerprint1", "user", 86400
            )

        self.assertEqual("silence1", silence_id1)
        fake_update_silence.assert_called_once_with(
            "fingerprint1", "user", 86400, force=False
        )
        fake_create_silence.assert_not_called()

    @freeze_time(datetime.utcfromtimestamp(0))
    async def test_create_silence_without_duration(self) -> None:
        fake_cache = Mock(spec=Cache)
        fake_cache.get.return_value = None

        async with FakeAlertmanagerServerWithoutSilence() as fake_alertmanager_server:
            port = fake_alertmanager_server.port
            alertmanager_client = AlertmanagerClient(
                f"http://localhost:{port}", fake_cache
            )
            async with aiotools.closing_async(alertmanager_client):
                silence_id = await alertmanager_client.create_silence(
                    "fingerprint1", "user"
                )
                silence = await alertmanager_client.get_silence("silence1")

        self.assertEqual("silence1", silence_id)
        self.assertEqual(
            {
                "id": "silence1",
                "status": {"state": "active"},
                "matchers": [
                    {
                        "name": "alertname",
                        "value": "alert1",
                        "isRegex": False,
                        "isEqual": True,
                    }
                ],
                "createdBy": "user",
                "startsAt": "1970-01-01T00:00:00",
                "endsAt": "1970-01-01T03:00:00",
                "comment": "Acknowledge alert from Matrix",
            },
            silence,
        )
        fake_cache.set.assert_called_once_with("fingerprint1", "silence1", expire=None)

    @freeze_time(datetime.utcfromtimestamp(0))
    async def test_update_silence_without_duration(self) -> None:
        fake_cache = FakeCache()

        async with FakeAlertmanagerServerWithoutSilence() as fake_alertmanager_server:
            port = fake_alertmanager_server.port
            alertmanager_client = AlertmanagerClient(
                f"http://localhost:{port}", fake_cache
            )
            async with aiotools.closing_async(alertmanager_client):
                silence1_id = await alertmanager_client.create_silence(
                    "fingerprint1", "user"
                )
                silence2_id = await alertmanager_client.update_silence("fingerprint1")
                silence = await alertmanager_client.get_silence("silence2")

        self.assertEqual("silence1", silence1_id)
        self.assertEqual("silence2", silence2_id)
        self.assertEqual(
            {
                "id": "silence2",
                "status": {"state": "active"},
                "matchers": [
                    {
                        "name": "alertname",
                        "value": "alert1",
                        "isRegex": False,
                        "isEqual": True,
                    }
                ],
                "createdBy": "user",
                "startsAt": "1970-01-01T00:00:00",
                "endsAt": "1970-01-01T03:00:00",
                "comment": "Acknowledge alert from Matrix",
            },
            silence,
        )
        self.assertEqual({"fingerprint1": ("silence2", None)}, fake_cache.cache)

    @patch.object(matrix_alertbot.alertmanager.AlertmanagerClient, "update_silence")
    @patch.object(matrix_alertbot.alertmanager.AlertmanagerClient, "create_silence")
    async def test_create_or_update_silence_without_duration_and_silence_not_found(
        self, fake_create_silence: Mock, fake_update_silence: Mock
    ) -> None:
        fake_cache = Mock(spec=Cache)
        fake_update_silence.side_effect = update_silence_raise_silence_not_found
        fake_create_silence.return_value = "silence1"

        alertmanager_client = AlertmanagerClient("http://localhost", fake_cache)
        async with aiotools.closing_async(alertmanager_client):
            silence_id1 = await alertmanager_client.create_or_update_silence(
                "fingerprint1", "user"
            )

        self.assertEqual("silence1", silence_id1)
        fake_update_silence.assert_called_once_with(
            "fingerprint1", "user", None, force=False
        )
        fake_create_silence.assert_called_once_with("fingerprint1", "user", None)

    @patch.object(matrix_alertbot.alertmanager.AlertmanagerClient, "update_silence")
    @patch.object(matrix_alertbot.alertmanager.AlertmanagerClient, "create_silence")
    async def test_create_or_update_silence_without_duration_and_silence_found(
        self, fake_create_silence: Mock, fake_update_silence: Mock
    ) -> None:
        fake_cache = Mock(spec=Cache)
        fake_update_silence.return_value = "silence1"

        alertmanager_client = AlertmanagerClient("http://localhost", fake_cache)
        async with aiotools.closing_async(alertmanager_client):
            silence_id1 = await alertmanager_client.create_or_update_silence(
                "fingerprint1", "user"
            )

        self.assertEqual("silence1", silence_id1)
        fake_update_silence.assert_called_once_with(
            "fingerprint1", "user", None, force=False
        )
        fake_create_silence.assert_not_called()

    @freeze_time(datetime.utcfromtimestamp(0))
    async def test_create_silence_with_max_duration(self) -> None:
        fake_cache = Mock(spec=Cache)
        fake_cache.get.return_value = None
        async with FakeAlertmanagerServerWithoutSilence() as fake_alertmanager_server:
            port = fake_alertmanager_server.port
            alertmanager_client = AlertmanagerClient(
                f"http://localhost:{port}", fake_cache
            )
            async with aiotools.closing_async(alertmanager_client):
                silence_id = await alertmanager_client.create_silence(
                    "fingerprint1", "user", int(timedelta.max.total_seconds()) + 1
                )
                silence = await alertmanager_client.get_silence("silence1")

        self.assertEqual("silence1", silence_id)
        self.assertEqual(
            {
                "id": "silence1",
                "status": {"state": "active"},
                "matchers": [
                    {
                        "name": "alertname",
                        "value": "alert1",
                        "isRegex": False,
                        "isEqual": True,
                    }
                ],
                "createdBy": "user",
                "startsAt": "1970-01-01T00:00:00",
                "endsAt": "1980-01-01T00:00:00",
                "comment": "Acknowledge alert from Matrix",
            },
            silence,
        )

    async def test_create_silence_raise_alert_not_found(self) -> None:
        fake_cache = Mock(spec=Cache)
        fake_cache.get.return_value = None

        async with FakeAlertmanagerServerWithoutAlert() as fake_alertmanager_server:
            port = fake_alertmanager_server.port
            alertmanager_client = AlertmanagerClient(
                f"http://localhost:{port}", fake_cache
            )
            async with aiotools.closing_async(alertmanager_client):
                with self.assertRaises(AlertNotFoundError):
                    await alertmanager_client.create_silence("fingerprint1", "user")

    async def test_create_silence_raise_alertmanager_error(self) -> None:
        fake_cache = Mock(spec=Cache)
        fake_cache.get.return_value = None

        async with FakeAlertmanagerServerWithErrorCreateSilence() as fake_alertmanager_server:
            port = fake_alertmanager_server.port
            alertmanager_client = AlertmanagerClient(
                f"http://localhost:{port}", fake_cache
            )
            async with aiotools.closing_async(alertmanager_client):
                await alertmanager_client.get_alert("fingerprint1")

                with self.assertRaises(AlertmanagerServerError):
                    await alertmanager_client.create_silence("fingerprint1", "user")

    async def test_update_silence_raise_silence_not_found(self) -> None:
        fake_cache = FakeCache({"fingerprint1": ("silence1", None)})

        async with FakeAlertmanagerServerWithoutSilence() as fake_alertmanager_server:
            port = fake_alertmanager_server.port
            alertmanager_client = AlertmanagerClient(
                f"http://localhost:{port}", fake_cache
            )
            async with aiotools.closing_async(alertmanager_client):
                with self.assertRaises(SilenceNotFoundError):
                    await alertmanager_client.update_silence("fingerprint1")
                with self.assertRaises(SilenceNotFoundError):
                    await alertmanager_client.update_silence("fingerprint2")

    async def test_update_silence_raise_silence_extend_error(self) -> None:
        fake_cache = FakeCache({"fingerprint1": ("silence1", 86400)})

        async with FakeAlertmanagerServerWithoutAlert() as fake_alertmanager_server:
            port = fake_alertmanager_server.port
            alertmanager_client = AlertmanagerClient(
                f"http://localhost:{port}", fake_cache
            )
            async with aiotools.closing_async(alertmanager_client):
                with self.assertRaises(SilenceExtendError):
                    await alertmanager_client.update_silence("fingerprint1")

    async def test_update_silence_raise_alertmanager_error(self) -> None:
        fake_cache = FakeCache({"fingerprint1": ("silence1", None)})

        async with FakeAlertmanagerServerWithErrorCreateSilence() as fake_alertmanager_server:
            port = fake_alertmanager_server.port
            alertmanager_client = AlertmanagerClient(
                f"http://localhost:{port}", fake_cache
            )
            async with aiotools.closing_async(alertmanager_client):
                await alertmanager_client.get_alert("fingerprint1")

                with self.assertRaises(AlertmanagerServerError):
                    await alertmanager_client.update_silence("fingerprint1")

    async def test_delete_silence(self) -> None:
        fake_cache = Mock(spec=Cache)

        async with FakeAlertmanagerServer() as fake_alertmanager_server:
            port = fake_alertmanager_server.port
            alertmanager_client = AlertmanagerClient(
                f"http://localhost:{port}", fake_cache
            )
            async with aiotools.closing_async(alertmanager_client):
                await alertmanager_client.delete_silence("silence1")
                silences = await alertmanager_client.get_silences()

        self.assertEqual(
            [
                {
                    "id": "silence2",
                    "createdBy": "user2",
                    "status": {"state": "expired"},
                    "matchers": [],
                }
            ],
            silences,
        )

    async def test_delete_silence_raise_silence_expired(self) -> None:
        fake_cache = Mock(spec=Cache)

        async with FakeAlertmanagerServer() as fake_alertmanager_server:
            port = fake_alertmanager_server.port
            alertmanager_client = AlertmanagerClient(
                f"http://localhost:{port}", fake_cache
            )
            async with aiotools.closing_async(alertmanager_client):
                with self.assertRaises(SilenceExpiredError):
                    await alertmanager_client.delete_silence("silence2")
                silences = await alertmanager_client.get_silences()

        self.assertEqual(2, len(silences))

    async def test_delete_silence_raise_alertmanager_error(self) -> None:
        fake_cache = Mock(spec=Cache)

        async with FakeAlertmanagerServerWithErrorDeleteSilence() as fake_alertmanager_server:
            port = fake_alertmanager_server.port
            alertmanager_client = AlertmanagerClient(
                f"http://localhost:{port}", fake_cache
            )
            async with aiotools.closing_async(alertmanager_client):
                await alertmanager_client.get_alert("fingerprint1")

                with self.assertRaises(AlertmanagerServerError):
                    await alertmanager_client.delete_silence("silence1")

    async def test_find_alert_happy(self) -> None:
        fake_cache = Mock(spec=Cache)

        alertmanager_client = AlertmanagerClient("http://localhost", fake_cache)
        alert = alertmanager_client._find_alert(
            "fingerprint1", [{"fingerprint": "fingerprint1"}]
        )
        self.assertEqual({"fingerprint": "fingerprint1"}, alert)

    async def test_find_alert_raise_alert_not_found(self) -> None:
        fake_cache = Mock(spec=Cache)

        alertmanager_client = AlertmanagerClient("http://localhost", fake_cache)

        with self.assertRaises(AlertNotFoundError):
            alertmanager_client._find_alert("fingerprint1", [])

        with self.assertRaises(AlertNotFoundError):
            alertmanager_client._find_alert(
                "fingerprint2", [{"fingerprint": "fingerprint1"}]
            )

    async def test_find_silence_happy(self) -> None:
        fake_cache = Mock(spec=Cache)

        alertmanager_client = AlertmanagerClient("http://localhost", fake_cache)
        silence = alertmanager_client._find_silence("silence1", [{"id": "silence1"}])
        self.assertEqual({"id": "silence1"}, silence)

    async def test_find_silence_raise_silence_not_found(self) -> None:
        fake_cache = Mock(spec=Cache)

        alertmanager_client = AlertmanagerClient("http://localhost", fake_cache)

        with self.assertRaises(SilenceNotFoundError):
            alertmanager_client._find_silence("silence1", [])

        with self.assertRaises(SilenceNotFoundError):
            alertmanager_client._find_silence("silence2", [{"id": "silence1"}])


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
import asyncio
import json
import logging
import os
import sys
from asyncio import TimeoutError

from aiohttp import ClientConnectionError, ServerDisconnectedError
from diskcache import Cache
from nio import (
    AsyncClient,
    AsyncClientConfig,
    InviteMemberEvent,
    KeyVerificationCancel,
    KeyVerificationKey,
    KeyVerificationMac,
    KeyVerificationStart,
    LocalProtocolError,
    LoginError,
    MegolmEvent,
    RedactionEvent,
    RoomMessageText,
    UnknownEvent,
)

from matrix_alertbot.alertmanager import AlertmanagerClient
from matrix_alertbot.callback import Callbacks
from matrix_alertbot.config import Config
from matrix_alertbot.webhook import Webhook

logger = logging.getLogger(__name__)


def create_matrix_client(config: Config) -> AsyncClient:
    # Configuration options for the AsyncClient
    try:
        matrix_client_config = AsyncClientConfig(
            max_limit_exceeded=0,
            max_timeouts=0,
            store_sync_tokens=True,
            encryption_enabled=True,
        )
    except ImportWarning as e:
        logger.warning(e)
        matrix_client_config = AsyncClientConfig(
            max_limit_exceeded=0,
            max_timeouts=0,
            store_sync_tokens=True,
            encryption_enabled=False,
        )

    # Load credentials from a previous session
    if os.path.exists(config.user_token_file):
        with open(config.user_token_file, "r") as ifd:
            credentials = json.load(ifd)
            config.user_token = credentials["access_token"]
            config.device_id = credentials["device_id"]

    # Initialize the matrix client based on stored credentials
    matrix_client = AsyncClient(
        config.homeserver_url,
        config.user_id,
        device_id=config.device_id,
        store_path=config.store_dir,
        config=matrix_client_config,
    )

    return matrix_client


async def start_matrix_client(
    matrix_client: AsyncClient, cache: Cache, config: Config
) -> bool:
    # Keep trying to reconnect on failure (with some time in-between)
    while True:
        try:
            if config.device_id and config.user_token:
                matrix_client.restore_login(
                    user_id=config.user_id,
                    device_id=config.device_id,
                    access_token=config.user_token,
                )

                # Sync encryption keys with the server
                if matrix_client.should_upload_keys:
                    await matrix_client.keys_upload()
            else:
                # Try to login with the configured username/password
                try:
                    login_response = await matrix_client.login(
                        password=config.user_password,
                        device_name=config.device_name,
                    )

                    # Check if login failed
                    if type(login_response) == LoginError:
                        logger.error("Failed to login: %s", login_response.message)
                        return False
                except LocalProtocolError as e:
                    # There's an edge case here where the user hasn't installed the correct C
                    # dependencies. In that case, a LocalProtocolError is raised on login.
                    logger.fatal(
                        "Failed to login. Have you installed the correct dependencies? "
                        "https://github.com/poljar/matrix-nio#installation "
                        "Error: %s",
                        e,
                    )
                    return False

                # Save user's access token and device ID
                # See https://stackoverflow.com/a/45368120
                user_token_fd = os.open(
                    config.user_token_file,
                    flags=os.O_CREAT | os.O_WRONLY | os.O_TRUNC,
                    mode=0o640,
                )
                with os.fdopen(user_token_fd, "w") as ofd:
                    json.dump(
                        {
                            "device_id": login_response.device_id,
                            "access_token": login_response.access_token,
                        },
                        ofd,
                    )

                # Login succeeded!

            logger.info(f"Logged in as {config.user_id}")
            await matrix_client.sync_forever(timeout=30000, full_state=True)
        except (ClientConnectionError, ServerDisconnectedError, TimeoutError):
            logger.warning("Unable to connect to homeserver, retrying in 15s...")

            # Sleep so we don't bombard the server with login requests
            await asyncio.sleep(15)
        finally:
            await matrix_client.close()


def main() -> None:
    """The first function that is run when starting the bot"""

    # Read user-configured options from a config file.
    # A different config file path can be specified as the first command line argument
    if len(sys.argv) > 1:
        config_path = sys.argv[1]
    else:
        config_path = "config.yaml"

    # Read the parsed config file and create a Config object
    config = Config(config_path)

    matrix_client = create_matrix_client(config)

    # Configure the cache
    cache = Cache(config.cache_dir)

    # Configure Alertmanager client
    alertmanager_client = AlertmanagerClient(config.alertmanager_url, cache)

    # Set up event callbacks
    callbacks = Callbacks(matrix_client, alertmanager_client, cache, config)
    matrix_client.add_event_callback(callbacks.message, (RoomMessageText,))
    matrix_client.add_event_callback(
        callbacks.invite_event_filtered_callback, (InviteMemberEvent,)
    )
    matrix_client.add_event_callback(callbacks.decryption_failure, (MegolmEvent,))
    matrix_client.add_event_callback(callbacks.unknown, (UnknownEvent,))
    matrix_client.add_event_callback(callbacks.redaction, (RedactionEvent,))
    matrix_client.add_to_device_callback(
        callbacks.key_verification_start, (KeyVerificationStart,)
    )
    matrix_client.add_to_device_callback(
        callbacks.key_verification_cancel, (KeyVerificationCancel,)
    )
    matrix_client.add_to_device_callback(
        callbacks.key_verification_confirm, (KeyVerificationKey,)
    )
    matrix_client.add_to_device_callback(
        callbacks.key_verification_end, (KeyVerificationMac,)
    )
    # Configure webhook server
    webhook_server = Webhook(matrix_client, alertmanager_client, cache, config)

    loop = asyncio.get_event_loop()
    loop.create_task(webhook_server.start())
    loop.create_task(start_matrix_client(matrix_client, cache, config))

    try:
        loop.run_forever()
    except Exception as e:
        logger.error(e)
    finally:
        loop.run_until_complete(webhook_server.close())
        loop.run_until_complete(alertmanager_client.close())
        loop.run_until_complete(matrix_client.close())
        cache.close()



from __future__ import annotations

import logging
from datetime import datetime, timedelta
from typing import Dict, List, Optional

import aiohttp
from aiohttp import ClientError
from aiohttp_prometheus_exporter.trace import PrometheusTraceConfig
from diskcache import Cache

from matrix_alertbot.errors import (
    AlertmanagerServerError,
    AlertNotFoundError,
    SilenceExpiredError,
    SilenceExtendError,
    SilenceNotFoundError,
)

DEFAULT_DURATION = timedelta(hours=3)
MAX_DURATION = timedelta(days=3652)


logger = logging.getLogger(__name__)


class AlertmanagerClient:
    def __init__(self, url: str, cache: Cache) -> None:
        self.api_url = f"{url}/api/v2"
        self.cache = cache
        self.session = aiohttp.ClientSession(trace_configs=[PrometheusTraceConfig()])

    async def close(self) -> None:
        await self.session.close()

    async def get_alerts(self) -> List[Dict]:
        try:
            async with self.session.get(f"{self.api_url}/alerts") as response:
                response.raise_for_status()
                return await response.json()
        except ClientError as e:
            raise AlertmanagerServerError(
                "Cannot fetch alerts from Alertmanager"
            ) from e

    async def get_alert(self, fingerprint: str) -> Dict:
        logger.debug(f"Fetching details for alert with fingerprint {fingerprint}")
        alerts = await self.get_alerts()
        return self._find_alert(fingerprint, alerts)

    async def get_silences(self) -> List[Dict]:
        try:
            async with self.session.get(f"{self.api_url}/silences") as response:
                response.raise_for_status()
                return await response.json()
        except ClientError as e:
            raise AlertmanagerServerError(
                "Cannot fetch silences from Alertmanager"
            ) from e

    async def get_silence(self, silence_id: str) -> Dict:
        logger.debug(f"Fetching details for silence with ID {silence_id}")
        silences = await self.get_silences()
        return self._find_silence(silence_id, silences)

    async def create_silence(
        self,
        fingerprint: str,
        user: str,
        duration_seconds: Optional[int] = None,
    ) -> str:
        alert = await self.get_alert(fingerprint)

        logger.debug(f"Creating silence for alert with fingerprint {fingerprint}")

        silence_matchers = [
            {"name": label, "value": value, "isRegex": False, "isEqual": True}
            for label, value in alert["labels"].items()
        ]

        return await self._create_or_update_silence(
            fingerprint, silence_matchers, user, duration_seconds
        )

    async def update_silence(
        self,
        fingerprint: str,
        user: Optional[str] = None,
        duration_seconds: Optional[int] = None,
        *,
        force: bool = False,
    ) -> str:
        logger.debug(
            f"Reading silence for alert with fingerprint {fingerprint} from cache"
        )
        try:
            silence_id: Optional[str]
            expire_time: Optional[int]
            silence_id, expire_time = self.cache.get(fingerprint, expire_time=True)
        except TypeError:
            silence_id = None

        if silence_id is None:
            raise SilenceNotFoundError(
                f"Cannot find silence for alert with fingerprint {fingerprint} in cache."
            )

        logger.debug(
            f"Updating silence with ID {silence_id} for alert with fingerprint {fingerprint}"
        )

        # If silence in cache had a duration, and the new silence doesn't have a duration
        # then we cannot update this silence.
        if not force and duration_seconds is None and expire_time is not None:
            raise SilenceExtendError(
                f"Cannot extend silence ID {silence_id} with static duration."
            )

        silence = await self.get_silence(silence_id)
        if user is None:
            user = silence["createdBy"]
        silence_matchers = silence["matchers"]

        return await self._create_or_update_silence(
            fingerprint, silence_matchers, user, duration_seconds, silence_id
        )

    async def create_or_update_silence(
        self,
        fingerprint: str,
        user: str,
        duration_seconds: Optional[int] = None,
        *,
        force: bool = False,
    ) -> str:
        try:
            silence_id = await self.update_silence(
                fingerprint, user, duration_seconds, force=force
            )
        except SilenceNotFoundError:
            silence_id = await self.create_silence(fingerprint, user, duration_seconds)
        return silence_id

    async def _create_or_update_silence(
        self,
        fingerprint: str,
        silence_matchers: List,
        user: str,
        duration_seconds: Optional[int] = None,
        silence_id: Optional[str] = None,
    ) -> str:
        if duration_seconds is None:
            duration_delta = DEFAULT_DURATION
        elif duration_seconds > MAX_DURATION.total_seconds():
            duration_delta = MAX_DURATION
        else:
            duration_delta = timedelta(seconds=duration_seconds)
        start_time = datetime.now()
        end_time = start_time + duration_delta

        silence = {
            "id": silence_id,
            "matchers": silence_matchers,
            "startsAt": start_time.isoformat(),
            "endsAt": end_time.isoformat(),
            "createdBy": user,
            "comment": "Acknowledge alert from Matrix",
        }

        try:
            async with self.session.post(
                f"{self.api_url}/silences", json=silence
            ) as response:
                response.raise_for_status()
                data = await response.json()
        except ClientError as e:
            raise AlertmanagerServerError(
                f"Cannot create silence for alert fingerprint {fingerprint}"
            ) from e

        self.cache.set(fingerprint, data["silenceID"], expire=duration_seconds)

        return data["silenceID"]

    async def delete_silence(self, silence_id: str) -> None:
        silence = await self.get_silence(silence_id)

        silence_state = silence["status"]["state"]
        if silence_state == "expired":
            raise SilenceExpiredError(
                f"Cannot delete already expired silence with ID {silence_id}"
            )

        try:
            async with self.session.delete(
                f"{self.api_url}/silence/{silence_id}"
            ) as response:
                response.raise_for_status()
        except ClientError as e:
            raise AlertmanagerServerError(
                f"Cannot delete silence with ID {silence_id}"
            ) from e

    @staticmethod
    def _find_alert(fingerprint: str, alerts: List[Dict]) -> Dict:
        for alert in alerts:
            if alert["fingerprint"] == fingerprint:
                return alert
        raise AlertNotFoundError(f"Cannot find alert with fingerprint {fingerprint}")

    @staticmethod
    def _find_silence(silence_id: str, silences: List[Dict]) -> Dict:
        for silence in silences:
            if silence["id"] == silence_id:
                return silence
        raise SilenceNotFoundError(f"Cannot find silence with ID {silence_id}")



from __future__ import annotations

import logging
from typing import Dict, Optional

from jinja2 import (
    BaseLoader,
    ChoiceLoader,
    Environment,
    FileSystemLoader,
    PackageLoader,
)

logger = logging.getLogger(__name__)


class Alert:
    EMOJIS = {"critical": "🔥", "warning": "⚠️", "resolved": "🥦"}
    COLORS = {"critical": "dc3545", "warning": "ffc107", "resolved": "33cc33"}

    def __init__(
        self,
        fingerprint: str,
        url: str,
        labels: Dict[str, str],
        annotations: Dict[str, str],
        firing: bool = True,
    ):
        self.fingerprint = fingerprint
        self.url = url
        self.firing = firing

        self.labels = labels
        self.annotations = annotations
        self.description = annotations["description"]

        if self.firing:
            self.status = self.labels["severity"]
        else:
            self.status = "resolved"

    @staticmethod
    def from_dict(data: Dict) -> Alert:
        return Alert(
            fingerprint=data["fingerprint"],
            url=data["generatorURL"],
            firing=data["status"] == "firing",
            labels=data["labels"],
            annotations=data["annotations"],
        )

    @property
    def emoji(self) -> str:
        return self.EMOJIS[self.status]

    @property
    def color(self) -> str:
        return self.COLORS[self.status]


class AlertRenderer:
    def __init__(self, template_dir: Optional[str] = None) -> None:
        loader: BaseLoader = PackageLoader("matrix_alertbot", "resources/templates")
        if template_dir is not None:
            loader = ChoiceLoader([FileSystemLoader(template_dir), loader])
        env = Environment(loader=loader, autoescape=True)

        self.html_template = env.get_template("alert.html.j2")
        self.text_template = env.get_template("alert.txt.j2")

    def render(self, alert: Alert, html: bool = True) -> str:
        if html:
            template = self.html_template
        else:
            template = self.text_template
        return template.render(vars(alert), color=alert.color, emoji=alert.emoji)



import logging
from typing import Optional, Tuple

import pytimeparse2
from diskcache import Cache
from nio import AsyncClient, MatrixRoom

from matrix_alertbot.alertmanager import AlertmanagerClient
from matrix_alertbot.chat_functions import send_text_to_room
from matrix_alertbot.config import Config
from matrix_alertbot.errors import (
    AlertmanagerError,
    AlertNotFoundError,
    SilenceExpiredError,
    SilenceNotFoundError,
)

logger = logging.getLogger(__name__)


class BaseCommand:
    def __init__(
        self,
        matrix_client: AsyncClient,
        cache: Cache,
        alertmanager_client: AlertmanagerClient,
        config: Config,
        room: MatrixRoom,
        sender: str,
        event_id: str,
        args: Tuple[str, ...] = None,
    ) -> None:
        """A command made by a user.

        Args:
            client: The client to communicate with Matrix.

            cache: Bot cache.

            alertmanager: The client to communicate with Alertmanager.

            config: Bot configuration parameters.

            cmd: The command and arguments.

            room: The room the command was sent in.

            sender: The sender of the event

            event_id: The ID of the event describing the command.
        """
        self.matrix_client = matrix_client
        self.cache = cache
        self.alertmanager_client = alertmanager_client
        self.config = config
        self.room = room
        self.sender = sender
        self.event_id = event_id

        if args is not None:
            self.args = args
        else:
            self.args = ()

    async def process(self) -> None:
        raise NotImplementedError


class BaseAlertCommand(BaseCommand):
    def __init__(
        self,
        client: AsyncClient,
        cache: Cache,
        alertmanager: AlertmanagerClient,
        config: Config,
        room: MatrixRoom,
        sender: str,
        event_id: str,
        reacted_to_event_id: str,
        args: Tuple[str, ...] = None,
    ) -> None:
        super().__init__(
            client, cache, alertmanager, config, room, sender, event_id, args
        )

        self.reacted_to_event_id = reacted_to_event_id


class AckAlertCommand(BaseAlertCommand):
    async def process(self) -> None:
        """Acknowledge an alert and silence it for a certain duration in Alertmanager"""
        durations = self.args
        if len(durations) > 0:
            duration = " ".join(durations)
            logger.debug(f"Receiving a command to create a silence for {duration}.")

            duration_seconds = pytimeparse2.parse(duration)
            if duration_seconds is None:
                logger.error(f"Unable to create silence: Invalid duration '{duration}'")
                await send_text_to_room(
                    self.matrix_client,
                    self.room.room_id,
                    f"I tried really hard, but I can't convert the duration '{duration}' to a number of seconds.",
                )
                return
            elif duration_seconds < 0:
                logger.error(
                    f"Unable to create silence: Duration must be positive, got '{duration}'"
                )
                await send_text_to_room(
                    self.matrix_client,
                    self.room.room_id,
                    "I can't create a silence with a negative duration!",
                )
                return
        else:
            duration_seconds = None
            logger.debug(
                "Receiving a command to create a silence for an indefinite period"
            )

        logger.debug(
            f"Reading alert fingerprint for event {self.reacted_to_event_id} from cache"
        )
        try:
            alert_fingerprint: str = self.cache[self.reacted_to_event_id]
        except KeyError:
            logger.error(
                f"Cannot find fingerprint for alert event {self.reacted_to_event_id} in cache"
            )
            return

        try:
            silence_id = await self.alertmanager_client.create_or_update_silence(
                alert_fingerprint,
                self.room.user_name(self.sender),
                duration_seconds,
                force=True,
            )
        except AlertNotFoundError as e:
            logger.warning(f"Unable to create silence: {e}")
            await send_text_to_room(
                self.matrix_client,
                self.room.room_id,
                f"Sorry, I couldn't create silence for alert with fingerprint {alert_fingerprint}: {e}",
            )
            return
        except AlertmanagerError as e:
            logger.exception(f"Unable to create silence: {e}", exc_info=e)
            await send_text_to_room(
                self.matrix_client,
                self.room.room_id,
                f"Sorry, I couldn't create silence for alert with fingerprint {alert_fingerprint} "
                f"because something went wrong with Alertmanager: {e}",
            )
            return

        self.cache.set(self.event_id, alert_fingerprint, expire=duration_seconds)

        await send_text_to_room(
            self.matrix_client,
            self.room.room_id,
            f"Created silence with ID {silence_id}.",
        )


class UnackAlertCommand(BaseAlertCommand):
    async def process(self) -> None:
        """Delete an alert's acknowledgement of an alert and remove corresponding silence in Alertmanager"""
        logger.debug("Receiving a command to delete a silence")

        logger.debug(
            f"Reading alert fingerprint for event {self.reacted_to_event_id} from cache."
        )
        try:
            alert_fingerprint: str = self.cache[self.reacted_to_event_id]
        except KeyError:
            logger.error(
                f"Cannot find fingerprint for alert event {self.reacted_to_event_id} in cache."
            )
            return
        logger.debug(f"Found alert fingerprint {alert_fingerprint} in cache.")

        logger.debug(
            f"Reading silence ID for alert fingerprint {alert_fingerprint} from cache."
        )
        try:
            silence_id: str = self.cache[alert_fingerprint]
        except KeyError:
            logger.error(
                f"Cannot find silence for alert fingerprint {alert_fingerprint} in cache"
            )
            return
        logger.debug(f"Found silence ID {silence_id} in cache.")

        logger.debug(
            f"Deleting silence with ID {silence_id} for alert with fingerprint {alert_fingerprint}"
        )

        try:
            await self.alertmanager_client.delete_silence(silence_id)
        except (SilenceNotFoundError, SilenceExpiredError) as e:
            logger.error(f"Unable to delete silence: {e}")
            await send_text_to_room(
                self.matrix_client,
                self.room.room_id,
                f"Sorry, I couldn't remove silence for alert with fingerprint {alert_fingerprint}: {e}",
            )
            return
        except AlertmanagerError as e:
            logger.exception(f"Unable to delete silence: {e}", exc_info=e)
            await send_text_to_room(
                self.matrix_client,
                self.room.room_id,
                f"Sorry, I couldn't remove silence for alert with fingerprint {alert_fingerprint} "
                f"because something went wrong with Alertmanager: {e}",
            )
            return

        self.cache.delete(alert_fingerprint)

        await send_text_to_room(
            self.matrix_client,
            self.room.room_id,
            f"Removed silence with ID {silence_id}.",
        )


class HelpCommand(BaseCommand):
    async def process(self) -> None:
        """Show the help text"""
        logger.debug(f"Displaying help to room {self.room.display_name}")
        if not self.args:
            text = (
                "Hello, I am a bot made with matrix-nio! Use `help commands` to view "
                "available commands."
            )
            await send_text_to_room(self.matrix_client, self.room.room_id, text)
            return

        topic = self.args[0]
        if topic == "rules":
            text = "These are the rules!"
        elif topic == "commands":
            text = "Available commands: ..."
        else:
            text = "Unknown help topic!"
        await send_text_to_room(self.matrix_client, self.room.room_id, text)


class UnknownCommand(BaseCommand):
    async def process(self) -> None:
        logger.debug(
            f"Sending unknown command response to room {self.room.display_name}"
        )
        await send_text_to_room(
            self.matrix_client,
            self.room.room_id,
            "Unknown command. Try the 'help' command for more information.",
        )


class CommandFactory:
    @staticmethod
    def create(
        cmd: str,
        client: AsyncClient,
        cache: Cache,
        alertmanager: AlertmanagerClient,
        config: Config,
        room: MatrixRoom,
        sender: str,
        event_id: str,
        reacted_to_event_id: Optional[str] = None,
    ) -> BaseCommand:
        args = tuple(cmd.split()[1:])

        if cmd.startswith("ack"):
            if reacted_to_event_id is None:
                raise TypeError("Alert command must be in reply to an alert event.")

            return AckAlertCommand(
                client,
                cache,
                alertmanager,
                config,
                room,
                sender,
                event_id,
                reacted_to_event_id,
                args,
            )
        elif cmd.startswith("unack") or cmd.startswith("nack"):
            if reacted_to_event_id is None:
                raise TypeError("Alert command must be in reply to an alert event.")

            return UnackAlertCommand(
                client,
                cache,
                alertmanager,
                config,
                room,
                sender,
                event_id,
                reacted_to_event_id,
                args,
            )
        elif cmd.startswith("help"):
            return HelpCommand(
                client, cache, alertmanager, config, room, sender, event_id, args
            )
        else:
            return UnknownCommand(
                client, cache, alertmanager, config, room, sender, event_id, args
            )



import os
import unittest
from typing import Dict

from matrix_alertbot.alert import Alert, AlertRenderer

TESTS_DIR = os.path.dirname(__file__)


class AlertTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.alert_dict: Dict = {
            "fingerprint": "fingerprint1",
            "generatorURL": "http://example.com",
            "status": "unknown",
            "labels": {"alertname": "alert1", "severity": "critical", "job": "job1"},
            "annotations": {"description": "some description"},
        }

    def test_create_firing_alert_from_dict(self) -> None:
        self.alert_dict["status"] = "firing"
        alert = Alert.from_dict(self.alert_dict)

        self.assertEqual("fingerprint1", alert.fingerprint)
        self.assertEqual("http://example.com", alert.url)
        self.assertTrue(alert.firing)
        self.assertEqual("critical", alert.status)
        self.assertDictEqual(
            {"alertname": "alert1", "severity": "critical", "job": "job1"}, alert.labels
        )
        self.assertDictEqual({"description": "some description"}, alert.annotations)

    def test_create_resolved_alert_from_dict(self) -> None:
        self.alert_dict["status"] = "resolved"
        alert = Alert.from_dict(self.alert_dict)

        self.assertEqual("resolved", alert.status)
        self.assertFalse(alert.firing)

    def test_create_unknown_alert_from_dict(self) -> None:
        alert = Alert.from_dict(self.alert_dict)

        self.assertEqual("resolved", alert.status)
        self.assertFalse(alert.firing)


class AlertRendererTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.alert_dict: Dict = {
            "fingerprint": "fingerprint1",
            "generatorURL": "http://example.com",
            "status": "unknown",
            "labels": {"alertname": "alert1", "severity": "critical", "job": "job1"},
            "annotations": {"description": "some description"},
        }
        self.renderer = AlertRenderer()

    def test_render_firing_critical_alert(self) -> None:
        self.alert_dict["status"] = "firing"
        alert = Alert.from_dict(self.alert_dict)
        alert.labels["severity"] = "critical"

        html = self.renderer.render(alert, html=True)
        self.assertEqual(
            '<font color="#dc3545">\n  <b>[🔥 CRITICAL]</b>\n</font> '
            '<a href="http://example.com">alert1</a>\n (job1)<br/>\n'
            "some description",
            html,
        )

        plaintext = self.renderer.render(alert, html=False)
        self.assertEqual("[🔥 CRITICAL] alert1: some description", plaintext)

    def test_render_firing_warning_alert(self) -> None:
        self.alert_dict["status"] = "firing"
        self.alert_dict["labels"]["severity"] = "warning"
        alert = Alert.from_dict(self.alert_dict)

        html = self.renderer.render(alert, html=True)
        self.assertEqual(
            '<font color="#ffc107">\n  <b>[⚠️ WARNING]</b>\n</font> '
            '<a href="http://example.com">alert1</a>\n (job1)<br/>\n'
            "some description",
            html,
        )

        plaintext = self.renderer.render(alert, html=False)
        self.assertEqual("[⚠️ WARNING] alert1: some description", plaintext)

    def test_render_firing_unknown_alert(self) -> None:
        self.alert_dict["status"] = "firing"
        self.alert_dict["labels"]["severity"] = "unknown"
        alert = Alert.from_dict(self.alert_dict)

        with self.assertRaisesRegex(KeyError, "unknown"):
            self.renderer.render(alert, html=True)

        with self.assertRaisesRegex(KeyError, "unknown"):
            self.renderer.render(alert, html=False)

    def test_render_resolved_alert(self) -> None:
        self.alert_dict["status"] = "resolved"
        alert = Alert.from_dict(self.alert_dict)

        html = self.renderer.render(alert, html=True)
        self.assertEqual(
            '<font color="#33cc33">\n  <b>[🥦 RESOLVED]</b>\n</font> '
            '<a href="http://example.com">alert1</a>\n (job1)<br/>\n'
            "some description",
            html,
        )

        plaintext = self.renderer.render(alert, html=False)
        self.assertEqual("[🥦 RESOLVED] alert1: some description", plaintext)

    def test_render_resolved_alert_without_job(self) -> None:
        self.alert_dict["status"] = "resolved"
        del self.alert_dict["labels"]["job"]
        alert = Alert.from_dict(self.alert_dict)

        html = self.renderer.render(alert, html=True)
        self.assertEqual(
            '<font color="#33cc33">\n  <b>[🥦 RESOLVED]</b>\n</font> '
            '<a href="http://example.com">alert1</a>\n<br/>\n'
            "some description",
            html,
        )

        plaintext = self.renderer.render(alert, html=False)
        self.assertEqual("[🥦 RESOLVED] alert1: some description", plaintext)

    def test_render_with_existing_filesystem_template(self) -> None:
        alert = Alert.from_dict(self.alert_dict)

        template_dir = os.path.join(TESTS_DIR, "resources/templates")
        renderer = AlertRenderer(template_dir)

        html = renderer.render(alert, html=True)
        self.assertEqual(
            "<b>hello world</b>",
            html,
        )

        plaintext = renderer.render(alert, html=False)
        self.assertEqual("hello world", plaintext)

    def test_render_with_inexistent_filesystem_template(self) -> None:
        self.alert_dict["status"] = "resolved"
        alert = Alert.from_dict(self.alert_dict)

        renderer = AlertRenderer(TESTS_DIR)
        html = renderer.render(alert, html=True)
        self.assertEqual(
            '<font color="#33cc33">\n  <b>[🥦 RESOLVED]</b>\n</font> '
            '<a href="http://example.com">alert1</a>\n (job1)<br/>\n'
            "some description",
            html,
        )

        plaintext = renderer.render(alert, html=False)
        self.assertEqual("[🥦 RESOLVED] alert1: some description", plaintext)


if __name__ == "__main__":
    unittest.main()



import unittest
from typing import Dict
from unittest.mock import Mock, call, patch

import aiohttp.test_utils
import nio
from aiohttp import web
from diskcache import Cache
from nio import LocalProtocolError, RoomSendResponse

import matrix_alertbot.webhook
from matrix_alertbot.alertmanager import AlertmanagerClient
from matrix_alertbot.config import Config
from matrix_alertbot.errors import (
    AlertmanagerError,
    SilenceExtendError,
    SilenceNotFoundError,
)
from matrix_alertbot.webhook import Webhook


def send_text_to_room_raise_error(
    client: nio.AsyncClient, room_id: str, plaintext: str, html: str, notice: bool
) -> RoomSendResponse:
    raise LocalProtocolError


def update_silence_raise_silence_not_found(fingerprint: str) -> str:
    raise SilenceNotFoundError


def update_silence_raise_silence_extend_error(fingerprint: str) -> str:
    raise SilenceExtendError


def update_silence_raise_alertmanager_error(fingerprint: str) -> str:
    raise AlertmanagerError


class WebhookApplicationTestCase(aiohttp.test_utils.AioHTTPTestCase):
    async def get_application(self) -> web.Application:
        self.fake_matrix_client = Mock(spec=nio.AsyncClient)
        self.fake_alertmanager_client = Mock(spec=AlertmanagerClient)
        self.fake_cache = Mock(spec=Cache)

        self.fake_room_id = "!abcdefg:example.com"

        self.fake_config = Mock(spec=Config)
        self.fake_config.port = aiohttp.test_utils.unused_port()
        self.fake_config.address = "localhost"
        self.fake_config.socket = "webhook.sock"
        self.fake_config.allowed_rooms = [self.fake_room_id]
        self.fake_config.cache_expire_time = 0
        self.fake_config.template_dir = None

        self.fake_alerts = {
            "alerts": [
                {
                    "fingerprint": "fingerprint1",
                    "generatorURL": "http://example.com/alert1",
                    "status": "firing",
                    "labels": {
                        "alertname": "alert1",
                        "severity": "critical",
                        "job": "job1",
                    },
                    "annotations": {"description": "some description1"},
                },
                {
                    "fingerprint": "fingerprint2",
                    "generatorURL": "http://example.com/alert2",
                    "status": "resolved",
                    "labels": {
                        "alertname": "alert2",
                        "severity": "warning",
                        "job": "job2",
                    },
                    "annotations": {"description": "some description2"},
                },
            ]
        }

        webhook = Webhook(
            self.fake_matrix_client,
            self.fake_alertmanager_client,
            self.fake_cache,
            self.fake_config,
        )
        return webhook.app

    @patch.object(matrix_alertbot.webhook, "send_text_to_room")
    async def test_post_alerts_with_silence_not_found(
        self, fake_send_text_to_room: Mock
    ) -> None:
        self.fake_alertmanager_client.update_silence.side_effect = (
            update_silence_raise_silence_not_found
        )

        data = self.fake_alerts
        async with self.client.request(
            "POST", f"/alerts/{self.fake_room_id}", json=data
        ) as response:
            self.assertEqual(200, response.status)

        self.fake_alertmanager_client.update_silence.assert_called_once_with(
            "fingerprint1"
        )
        self.assertEqual(2, fake_send_text_to_room.call_count)
        fake_send_text_to_room.assert_has_calls(
            [
                call(
                    self.fake_matrix_client,
                    self.fake_room_id,
                    "[🔥 CRITICAL] alert1: some description1",
                    '<font color="#dc3545">\n  <b>[🔥 CRITICAL]</b>\n</font> '
                    '<a href="http://example.com/alert1">alert1</a>\n (job1)<br/>\n'
                    "some description1",
                    notice=False,
                ),
                call(
                    self.fake_matrix_client,
                    self.fake_room_id,
                    "[🥦 RESOLVED] alert2: some description2",
                    '<font color="#33cc33">\n  <b>[🥦 RESOLVED]</b>\n</font> '
                    '<a href="http://example.com/alert2">alert2</a>\n (job2)<br/>\n'
                    "some description2",
                    notice=False,
                ),
            ]
        )
        self.fake_cache.set.assert_called_once_with(
            fake_send_text_to_room.return_value.event_id,
            "fingerprint1",
            expire=self.fake_config.cache_expire_time,
        )
        self.assertEqual(2, self.fake_cache.delete.call_count)
        self.fake_cache.delete.assert_has_calls(
            [call("fingerprint1"), call("fingerprint2")]
        )

    @patch.object(matrix_alertbot.webhook, "send_text_to_room")
    async def test_post_alerts_with_silence_extend_error(
        self, fake_send_text_to_room: Mock
    ) -> None:
        self.fake_alertmanager_client.update_silence.side_effect = (
            update_silence_raise_silence_extend_error
        )

        data = self.fake_alerts
        async with self.client.request(
            "POST", f"/alerts/{self.fake_room_id}", json=data
        ) as response:
            self.assertEqual(200, response.status)

        self.fake_alertmanager_client.update_silence.assert_called_once_with(
            "fingerprint1"
        )
        self.assertEqual(2, fake_send_text_to_room.call_count)
        fake_send_text_to_room.assert_has_calls(
            [
                call(
                    self.fake_matrix_client,
                    self.fake_room_id,
                    "[🔥 CRITICAL] alert1: some description1",
                    '<font color="#dc3545">\n  <b>[🔥 CRITICAL]</b>\n</font> '
                    '<a href="http://example.com/alert1">alert1</a>\n (job1)<br/>\n'
                    "some description1",
                    notice=False,
                ),
                call(
                    self.fake_matrix_client,
                    self.fake_room_id,
                    "[🥦 RESOLVED] alert2: some description2",
                    '<font color="#33cc33">\n  <b>[🥦 RESOLVED]</b>\n</font> '
                    '<a href="http://example.com/alert2">alert2</a>\n (job2)<br/>\n'
                    "some description2",
                    notice=False,
                ),
            ]
        )
        self.fake_cache.set.assert_called_once_with(
            fake_send_text_to_room.return_value.event_id,
            "fingerprint1",
            expire=self.fake_config.cache_expire_time,
        )
        self.fake_cache.delete.assert_called_once_with("fingerprint2")

    @patch.object(matrix_alertbot.webhook, "send_text_to_room")
    async def test_post_alerts_with_alertmanager_error(
        self, fake_send_text_to_room: Mock
    ) -> None:
        self.fake_alertmanager_client.update_silence.side_effect = (
            update_silence_raise_alertmanager_error
        )

        data = self.fake_alerts
        async with self.client.request(
            "POST", f"/alerts/{self.fake_room_id}", json=data
        ) as response:
            self.assertEqual(500, response.status)

        self.fake_alertmanager_client.update_silence.assert_called_once_with(
            "fingerprint1"
        )
        fake_send_text_to_room.assert_not_called()
        self.fake_cache.set.assert_not_called()
        self.fake_cache.delete.assert_not_called()

    @patch.object(matrix_alertbot.webhook, "send_text_to_room")
    async def test_post_alerts_with_existing_silence(
        self, fake_send_text_to_room: Mock
    ) -> None:
        self.fake_alertmanager_client.update_silence.return_value = "silence1"

        data = self.fake_alerts
        async with self.client.request(
            "POST", f"/alerts/{self.fake_room_id}", json=data
        ) as response:
            self.assertEqual(200, response.status)

        self.fake_alertmanager_client.update_silence.assert_called_once_with(
            "fingerprint1"
        )
        fake_send_text_to_room.assert_called_once_with(
            self.fake_matrix_client,
            self.fake_room_id,
            "[🥦 RESOLVED] alert2: some description2",
            '<font color="#33cc33">\n  <b>[🥦 RESOLVED]</b>\n</font> '
            '<a href="http://example.com/alert2">alert2</a>\n (job2)<br/>\n'
            "some description2",
            notice=False,
        )
        self.fake_cache.set.assert_not_called()
        self.fake_cache.delete.assert_called_once_with("fingerprint2")

    @patch.object(matrix_alertbot.webhook, "send_text_to_room")
    async def test_post_alerts_in_unauthorized_room(
        self, fake_send_text_to_room: Mock
    ) -> None:
        room_id = "!unauthorized_room@example.com"
        async with self.client.request(
            "POST", f"/alerts/{room_id}", json=self.fake_alerts
        ) as response:
            self.assertEqual(401, response.status)
            error_msg = await response.text()

        self.assertEqual(
            "Cannot send alerts to room ID !unauthorized_room@example.com.", error_msg
        )
        fake_send_text_to_room.assert_not_called()
        self.fake_cache.set.assert_not_called()
        self.fake_cache.delete.assert_not_called()

    @patch.object(matrix_alertbot.webhook, "send_text_to_room")
    async def test_post_alerts_with_empty_data(
        self, fake_send_text_to_room: Mock
    ) -> None:
        async with self.client.request(
            "POST", f"/alerts/{self.fake_room_id}", json={}
        ) as response:
            self.assertEqual(400, response.status)
            error_msg = await response.text()

        self.assertEqual("Data must contain 'alerts' key.", error_msg)
        fake_send_text_to_room.assert_not_called()
        self.fake_cache.set.assert_not_called()
        self.fake_cache.delete.assert_not_called()

    @patch.object(matrix_alertbot.webhook, "send_text_to_room")
    async def test_post_empty_alerts(self, fake_send_text_to_room: Mock) -> None:
        data: Dict = {"alerts": []}
        async with self.client.request(
            "POST", f"/alerts/{self.fake_room_id}", json=data
        ) as response:
            self.assertEqual(400, response.status)
            error_msg = await response.text()

        self.assertEqual("Alerts cannot be empty.", error_msg)
        fake_send_text_to_room.assert_not_called()
        self.fake_cache.set.assert_not_called()
        self.fake_cache.delete.assert_not_called()

    @patch.object(matrix_alertbot.webhook, "send_text_to_room")
    async def test_post_invalid_alerts(self, fake_send_text_to_room: Mock) -> None:
        data = {"alerts": "invalid"}
        async with self.client.request(
            "POST", f"/alerts/{self.fake_room_id}", json=data
        ) as response:
            self.assertEqual(400, response.status)
            error_msg = await response.text()

        self.assertEqual("Alerts must be a list, got 'str'.", error_msg)
        fake_send_text_to_room.assert_not_called()
        self.fake_cache.set.assert_not_called()
        self.fake_cache.delete.assert_not_called()

    @patch.object(matrix_alertbot.webhook, "send_text_to_room")
    async def test_post_alerts_with_empty_items(
        self, fake_send_text_to_room: Mock
    ) -> None:
        data: Dict = {"alerts": [{}]}
        async with self.client.request(
            "POST", f"/alerts/{self.fake_room_id}", json=data
        ) as response:
            self.assertEqual(400, response.status)
            error_msg = await response.text()

        self.assertEqual("Invalid alert: {}.", error_msg)
        fake_send_text_to_room.assert_not_called()
        self.fake_cache.set.assert_not_called()
        self.fake_cache.delete.assert_not_called()

    @patch.object(
        matrix_alertbot.webhook,
        "send_text_to_room",
        side_effect=send_text_to_room_raise_error,
    )
    async def test_post_alerts_raise_send_error(
        self, fake_send_text_to_room: Mock
    ) -> None:
        self.fake_alertmanager_client.update_silence.side_effect = (
            update_silence_raise_silence_not_found
        )

        data = self.fake_alerts
        async with self.client.request(
            "POST", f"/alerts/{self.fake_room_id}", json=data
        ) as response:
            self.assertEqual(500, response.status)
            error_msg = await response.text()

        self.assertEqual(
            "An error occured when sending alert with fingerprint 'fingerprint1' to Matrix room.",
            error_msg,
        )
        fake_send_text_to_room.assert_called_once()
        self.fake_cache.set.assert_not_called()
        self.fake_cache.delete.assert_called_once_with("fingerprint1")

    async def test_health(self) -> None:
        async with self.client.request("GET", "/health") as response:
            self.assertEqual(200, response.status)

    async def test_metrics(self) -> None:
        async with self.client.request("GET", "/metrics") as response:
            self.assertEqual(200, response.status)


class WebhookServerTestCase(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        self.fake_matrix_client = Mock(spec=nio.AsyncClient)
        self.fake_alertmanager_client = Mock(spec=AlertmanagerClient)
        self.fake_cache = Mock(spec=Cache)

        self.fake_config = Mock(spec=Config)
        self.fake_config.port = aiohttp.test_utils.unused_port()
        self.fake_config.address = "localhost"
        self.fake_config.socket = "webhook.sock"
        self.fake_config.cache_expire_time = 0
        self.fake_config.template_dir = None

    @patch.object(matrix_alertbot.webhook.web, "TCPSite", autospec=True)
    async def test_webhook_start_address_port(self, fake_tcp_site: Mock) -> None:
        webhook = Webhook(
            self.fake_matrix_client,
            self.fake_alertmanager_client,
            self.fake_cache,
            self.fake_config,
        )
        await webhook.start()

        fake_tcp_site.assert_called_once_with(
            webhook.runner, self.fake_config.address, self.fake_config.port
        )

        await webhook.close()

    @patch.object(matrix_alertbot.webhook.web, "UnixSite", autospec=True)
    async def test_webhook_start_unix_socket(self, fake_unix_site: Mock) -> None:
        self.fake_config.address = None
        self.fake_config.port = None

        webhook = Webhook(
            self.fake_matrix_client,
            self.fake_alertmanager_client,
            self.fake_cache,
            self.fake_config,
        )
        await webhook.start()

        fake_unix_site.assert_called_once_with(webhook.runner, self.fake_config.socket)

        await webhook.close()



import logging

from diskcache import Cache
from nio import (
    AsyncClient,
    InviteMemberEvent,
    JoinError,
    KeyVerificationCancel,
    KeyVerificationKey,
    KeyVerificationMac,
    KeyVerificationStart,
    LocalProtocolError,
    MatrixRoom,
    MegolmEvent,
    RedactionEvent,
    RoomGetEventError,
    RoomMessageText,
    SendRetryError,
    ToDeviceError,
    UnknownEvent,
)

from matrix_alertbot.alertmanager import AlertmanagerClient
from matrix_alertbot.chat_functions import strip_fallback
from matrix_alertbot.command import AckAlertCommand, CommandFactory, UnackAlertCommand
from matrix_alertbot.config import Config

logger = logging.getLogger(__name__)


class Callbacks:
    def __init__(
        self,
        matrix_client: AsyncClient,
        alertmanager_client: AlertmanagerClient,
        cache: Cache,
        config: Config,
    ):
        """
        Args:
            client: nio client used to interact with matrix.

            cache: Bot cache.

            alertmanager: Client used to interact with alertmanager.

            config: Bot configuration parameters.
        """
        self.matrix_client = matrix_client
        self.cache = cache
        self.alertmanager_client = alertmanager_client
        self.config = config
        self.command_prefix = config.command_prefix

    async def message(self, room: MatrixRoom, event: RoomMessageText) -> None:
        """Callback for when a message event is received

        Args:
            room: The room the event came from.

            event: The event defining the message.
        """
        # Ignore messages from ourselves
        if event.sender == self.matrix_client.user:
            return

        # Ignore messages from unauthorized room
        if room.room_id not in self.config.allowed_rooms:
            return

        # Extract the message text
        msg = strip_fallback(event.body)

        logger.debug(
            f"Bot message received for room {room.display_name} | "
            f"{room.user_name(event.sender)}: {msg}"
        )
        # Process as message if in a public room without command prefix
        has_command_prefix = msg.startswith(self.command_prefix)
        if not has_command_prefix:
            logger.debug(
                f"Cannot process message: Command prefix {self.command_prefix} not provided."
            )
            return

        source_content = event.source["content"]
        reacted_to_event_id = (
            source_content.get("m.relates_to", {})
            .get("m.in_reply_to", {})
            .get("event_id")
        )

        if reacted_to_event_id is not None:
            logger.debug(f"Command in reply to event ID {reacted_to_event_id}")

        # Remove the command prefix
        cmd = msg[len(self.command_prefix) :]
        try:
            command = CommandFactory.create(
                cmd,
                self.matrix_client,
                self.cache,
                self.alertmanager_client,
                self.config,
                room,
                event.sender,
                event.event_id,
                reacted_to_event_id,
            )
        except TypeError as e:
            logging.error(f"Cannot process command '{cmd}': {e}")
            return

        try:
            await command.process()
        except (SendRetryError, LocalProtocolError) as e:
            logger.exception(f"Unable to send message to {room.room_id}", exc_info=e)

    async def invite(self, room: MatrixRoom, event: InviteMemberEvent) -> None:
        """Callback for when an invite is received. Join the room specified in the invite.

        Args:
            room: The room that we are invited to.

            event: The invite event.
        """
        # Ignore invites from unauthorized room
        if room.room_id not in self.config.allowed_rooms:
            return

        logger.debug(f"Got invite to {room.room_id} from {event.sender}.")

        # Attempt to join 3 times before giving up
        for attempt in range(3):
            result = await self.matrix_client.join(room.room_id)
            if type(result) == JoinError:
                logger.error(
                    f"Error joining room {room.room_id} (attempt %d): %s",
                    attempt,
                    result.message,
                )
            else:
                break
        else:
            logger.error("Unable to join room: %s", room.room_id)

        # Successfully joined room
        logger.info(f"Joined {room.room_id}")

    async def invite_event_filtered_callback(
        self, room: MatrixRoom, event: InviteMemberEvent
    ) -> None:
        """
        Since the InviteMemberEvent is fired for every m.room.member state received
        in a sync response's `rooms.invite` section, we will receive some that are
        not actually our own invite event (such as the inviter's membership).
        This makes sure we only call `callbacks.invite` with our own invite events.
        """
        if event.state_key == self.matrix_client.user_id:
            # This is our own membership (invite) event
            await self.invite(room, event)

    async def _reaction(
        self, room: MatrixRoom, event: UnknownEvent, alert_event_id: str
    ) -> None:
        """A reaction was sent to one of our messages. Let's send a reply acknowledging it.

        Args:
            room: The room the reaction was sent in.

            event: The reaction event.

            reacted_to_id: The event ID that the reaction points to.
        """
        # Ignore reactions from unauthorized room
        if room.room_id not in self.config.allowed_rooms:
            return

        # Ignore reactions from ourselves
        if event.sender == self.matrix_client.user:
            return

        reaction = event.source.get("content", {}).get("m.relates_to", {}).get("key")
        logger.debug(f"Got reaction {reaction} to {room.room_id} from {event.sender}.")

        if reaction not in self.config.allowed_reactions:
            logger.warning(f"Uknown duration reaction {reaction}")
            return

        # Get the original event that was reacted to
        event_response = await self.matrix_client.room_get_event(
            room.room_id, alert_event_id
        )
        if isinstance(event_response, RoomGetEventError):
            logger.warning(
                f"Error getting event that was reacted to ({alert_event_id})"
            )
            return
        reacted_to_event = event_response.event

        # Only acknowledge reactions to events that we sent
        if reacted_to_event.sender != self.config.user_id:
            return

        # Send a message acknowledging the reaction
        command = AckAlertCommand(
            self.matrix_client,
            self.cache,
            self.alertmanager_client,
            self.config,
            room,
            event.sender,
            event.event_id,
            alert_event_id,
        )

        try:
            await command.process()
        except (SendRetryError, LocalProtocolError) as e:
            logger.exception(f"Unable to send message to {room.room_id}", exc_info=e)

    async def redaction(self, room: MatrixRoom, event: RedactionEvent) -> None:
        # Ignore events from unauthorized room
        if room.room_id not in self.config.allowed_rooms:
            return

        # Ignore redactions from ourselves
        if event.sender == self.matrix_client.user:
            return

        logger.debug(f"Received event to remove event ID {event.redacts}")

        command = UnackAlertCommand(
            self.matrix_client,
            self.cache,
            self.alertmanager_client,
            self.config,
            room,
            event.sender,
            event.event_id,
            event.redacts,
        )
        try:
            await command.process()
        except (SendRetryError, LocalProtocolError) as e:
            logger.exception(f"Unable to send message to {room.room_id}", exc_info=e)

    async def decryption_failure(self, room: MatrixRoom, event: MegolmEvent) -> None:
        """Callback for when an event fails to decrypt. Inform the user.

        Args:
            room: The room that the event that we were unable to decrypt is in.

            event: The encrypted event that we were unable to decrypt.
        """
        # Ignore events from unauthorized room
        if room.room_id not in self.config.allowed_rooms:
            return

        logger.error(
            f"Failed to decrypt event '{event.event_id}' in room '{room.room_id}'!"
            f"\n\n"
            f"Tip: try using a different device ID in your config file and restart."
            f"\n\n"
            f"If all else fails, delete your store directory and let the bot recreate "
            f"it (your reminders will NOT be deleted, but the bot may respond to existing "
            f"commands a second time)."
        )

    async def key_verification_start(self, event: KeyVerificationStart):
        """Callback for when somebody wants to verify our devices."""
        if "emoji" not in event.short_authentication_string:
            logger.error(
                f"Unable to use emoji verification with {event.sender} on device {event.from_device}."
            )
            return

        event_response = await self.matrix_client.accept_key_verification(
            event.transaction_id
        )
        if isinstance(event_response, ToDeviceError):
            logger.error(
                f"Unable to start key verification with {event.sender} on device {event.from_device}, got error: {event_response}."
            )
            return

        sas = self.matrix_client.key_verifications[event.transaction_id]

        todevice_msg = sas.share_key()
        event_response = await self.matrix_client.to_device(todevice_msg)
        if isinstance(event_response, ToDeviceError):
            logger.error(
                f"Unable to share key with {event.sender} on device {event.from_device}, got error: {event_response}."
            )
            return

    async def key_verification_cancel(self, event: KeyVerificationCancel):
        # There is no need to issue a
        # client.cancel_key_verification(tx_id, reject=False)
        # here. The SAS flow is already cancelled.
        # We only need to inform the user.
        logger.info(
            f"Verification has been cancelled by {event.sender} for reason: {event.reason}."
        )

    async def key_verification_confirm(self, event: KeyVerificationKey):
        sas = self.matrix_client.key_verifications[event.transaction_id]
        emoji_list, alt_text_list = zip(*sas.get_emoji())
        emoji_str = " ".join(emoji_list)
        alt_text_str = " ".join(alt_text_list)

        logger.info(
            f"Received request to verify emojis from {event.sender}: {emoji_str} ({alt_text_str})"
        )

        event_response = await self.matrix_client.confirm_short_auth_string(
            event.transaction_id
        )
        if isinstance(event_response, ToDeviceError):
            logger.error(
                f"Unable to confirm emoji verification with {event.sender}, got error: {event_response}."
            )

        # FIXME: We should allow manual cancel or reject
        # event_response = await self.matrix_client.cancel_key_verification(
        #     event.transaction_id, reject=True
        # )
        # if isinstance(event_response, ToDeviceError):
        #     logger.error(
        #         f"Unable to reject emoji verification with {event.sender}, got error: {event_response}."
        #     )
        #
        # event_response = await self.matrix_client.cancel_key_verification(
        #     event.transaction_id, reject=False
        # )
        # if isinstance(event_response, ToDeviceError):
        #     logger.error(
        #         f"Unable to cancel emoji verification with {event.sender}, got error: {event_response}."
        #     )

    async def key_verification_end(self, event: KeyVerificationMac):
        try:
            sas = self.matrix_client.key_verifications[event.transaction_id]
        except KeyError:
            logger.error(
                f"Unable to find transaction ID {event.transaction_id} sent by {event.sender}"
            )
            return

        try:
            todevice_msg = sas.get_mac()
        except LocalProtocolError as e:
            # e.g. it might have been cancelled by ourselves
            logger.warning(f"Unable to conclude verification with {event.sender}: {e}.")
            return

        event_response = await self.matrix_client.to_device(todevice_msg)
        if isinstance(event_response, ToDeviceError):
            logger.error(
                f"Unable to conclude verification with {event.sender}, got error: {event_response}."
            )
            return

        verified_devices = " ".join(sas.verified_devices)
        logger.info(
            f"Successfully verified devices from {event.sender}: {verified_devices}"
        )

    async def unknown(self, room: MatrixRoom, event: UnknownEvent) -> None:
        """Callback for when an event with a type that is unknown to matrix-nio is received.
        Currently this is used for reaction events, which are not yet part of a released
        matrix spec (and are thus unknown to nio).

        Args:
            room: The room the reaction was sent in.

            event: The event itself.
        """
        # Ignore events from unauthorized room
        if room.room_id not in self.config.allowed_rooms:
            return

        if event.type == "m.reaction":
            # Get the ID of the event this was a reaction to
            relation_dict = event.source.get("content", {}).get("m.relates_to", {})

            reacted_to_id = relation_dict.get("event_id")
            if reacted_to_id and relation_dict.get("rel_type") == "m.annotation":
                await self._reaction(room, event, reacted_to_id)
                return

        logger.debug(
            f"Got unknown event with type to {event.type} from {event.sender} in {room.room_id}."
        )



import unittest
from typing import Dict, Optional
from unittest.mock import MagicMock, Mock, call, patch

import nio
from diskcache import Cache

import matrix_alertbot.callback
import matrix_alertbot.command
from matrix_alertbot.alertmanager import AlertmanagerClient
from matrix_alertbot.command import (
    AckAlertCommand,
    CommandFactory,
    HelpCommand,
    UnackAlertCommand,
    UnknownCommand,
)
from matrix_alertbot.errors import (
    AlertmanagerError,
    AlertNotFoundError,
    SilenceNotFoundError,
)

from tests.utils import make_awaitable


def cache_get_item(key: str) -> str:
    return {
        "some alert event id": "fingerprint1",
        "fingerprint1": "silence1",
    }[key]


async def create_silence(
    fingerprint: str,
    user: str,
    duration_seconds: Optional[int] = None,
    *,
    force: bool = True,
) -> str:
    if fingerprint == "fingerprint1":
        return "silence1"
    elif fingerprint == "fingerprint2":
        return "silence2"
    raise AlertmanagerError


async def create_silence_raise_alertmanager_error(
    fingerprint: str,
    user: str,
    duration_seconds: Optional[int] = None,
    *,
    force: bool = True,
) -> str:
    if fingerprint == "fingerprint1":
        raise AlertmanagerError
    return "silence1"


async def create_silence_raise_alert_not_found_error(
    fingerprint: str,
    user: str,
    duration_seconds: Optional[int] = None,
    *,
    force: bool = True,
) -> str:
    if fingerprint == "fingerprint1":
        raise AlertNotFoundError
    return "silence1"


async def delete_silence_raise_alertmanager_error(silence_id: str) -> None:
    if silence_id == "silence1":
        raise AlertmanagerError


async def delete_silence_raise_silence_not_found_error(silence_id: str) -> None:
    if silence_id == "silence1":
        raise SilenceNotFoundError


class CommandTestCase(unittest.IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        # Create a Command object and give it some Mock'd objects to use
        self.fake_matrix_client = Mock(spec=nio.AsyncClient)
        self.fake_matrix_client.user = "@fake_user:example.com"
        # Pretend that attempting to send a message is always successful
        self.fake_matrix_client.room_send.return_value = make_awaitable()

        self.fake_cache = MagicMock(spec=Cache)
        self.fake_cache.__getitem__.side_effect = cache_get_item
        self.fake_cache.__contains__.return_value = True

        self.fake_alertmanager_client = Mock(spec=AlertmanagerClient)
        self.fake_alertmanager_client.create_or_update_silence.side_effect = (
            create_silence
        )

        # Create a fake room to play with
        self.fake_room = Mock(spec=nio.MatrixRoom)
        self.fake_room.room_id = "!abcdefg:example.com"
        self.fake_room.display_name = "Fake Room"
        self.fake_room.user_name.side_effect = lambda x: x

        self.fake_event_id = "some event id"
        self.fake_sender = "@some_other_fake_user:example.com"
        self.fake_alert_event_id = "some alert event id"

        # We don't spec config, as it doesn't currently have well defined attributes
        self.fake_config = Mock()
        self.fake_config.allowed_rooms = [self.fake_room.room_id]
        self.fake_config.command_prefix = "!alert "

    @patch.object(matrix_alertbot.command.AckAlertCommand, "process")
    async def test_process_ack_command(self, fake_ack: Mock) -> None:
        """Tests the callback for InviteMemberEvents"""
        # Tests that the bot attempts to join a room after being invited to it

        command = CommandFactory.create(
            "ack",
            self.fake_matrix_client,
            self.fake_cache,
            self.fake_alertmanager_client,
            self.fake_config,
            self.fake_room,
            self.fake_sender,
            self.fake_event_id,
            self.fake_alert_event_id,
        )
        await command.process()

        # Check that we attempted to process the command
        fake_ack.assert_called_once()

    @patch.object(matrix_alertbot.command.AckAlertCommand, "process")
    async def test_process_ack_with_duration_command(self, fake_ack: Mock) -> None:
        """Tests the callback for InviteMemberEvents"""
        # Tests that the bot attempts to join a room after being invited to it

        command = CommandFactory.create(
            "ack 1w 3d",
            self.fake_matrix_client,
            self.fake_cache,
            self.fake_alertmanager_client,
            self.fake_config,
            self.fake_room,
            self.fake_sender,
            self.fake_event_id,
            self.fake_alert_event_id,
        )
        await command.process()

        # Check that we attempted to process the command
        fake_ack.assert_called_once()

    @patch.object(matrix_alertbot.command.UnackAlertCommand, "process")
    async def test_process_unack_command(self, fake_unack: Mock) -> None:
        """Tests the callback for InviteMemberEvents"""
        # Tests that the bot attempts to join a room after being invited to it

        for unack_cmd in ("unack", "nack"):
            command = CommandFactory.create(
                unack_cmd,
                self.fake_matrix_client,
                self.fake_cache,
                self.fake_alertmanager_client,
                self.fake_config,
                self.fake_room,
                self.fake_sender,
                self.fake_event_id,
                self.fake_alert_event_id,
            )
            await command.process()

        # Check that we attempted to process the command
        fake_unack.assert_has_calls([call(), call()])

    @patch.object(matrix_alertbot.command.HelpCommand, "process")
    async def test_process_help_command(self, fake_help: Mock) -> None:
        """Tests the callback for InviteMemberEvents"""
        # Tests that the bot attempts to join a room after being invited to it

        command = CommandFactory.create(
            "help",
            self.fake_matrix_client,
            self.fake_cache,
            self.fake_alertmanager_client,
            self.fake_config,
            self.fake_room,
            self.fake_sender,
            self.fake_event_id,
        )
        await command.process()

        # Check that we attempted to process the command
        fake_help.assert_called_once()

    @patch.object(matrix_alertbot.command.UnknownCommand, "process")
    async def test_process_unknown_command(self, fake_unknown: Mock) -> None:
        """Tests the callback for InviteMemberEvents"""
        # Tests that the bot attempts to join a room after being invited to it

        command = CommandFactory.create(
            "",
            self.fake_matrix_client,
            self.fake_cache,
            self.fake_alertmanager_client,
            self.fake_config,
            self.fake_room,
            self.fake_sender,
            self.fake_event_id,
        )
        await command.process()

        # Check that we attempted to process the command
        fake_unknown.assert_called_once()

    @patch.object(matrix_alertbot.command, "send_text_to_room")
    async def test_ack_without_duration(self, fake_send_text_to_room: Mock) -> None:
        """Tests the callback for InviteMemberEvents"""
        # Tests that the bot attempts to join a room after being invited to it
        fake_cache_dict = {self.fake_alert_event_id: "fingerprint1"}

        self.fake_cache.__getitem__.side_effect = fake_cache_dict.__getitem__

        command = AckAlertCommand(
            self.fake_matrix_client,
            self.fake_cache,
            self.fake_alertmanager_client,
            self.fake_config,
            self.fake_room,
            self.fake_sender,
            self.fake_event_id,
            self.fake_alert_event_id,
        )
        await command.process()

        # Check that we attempted to create silences
        self.fake_alertmanager_client.create_or_update_silence.assert_called_once_with(
            "fingerprint1", self.fake_sender, None, force=True
        )
        fake_send_text_to_room.assert_called_once_with(
            self.fake_matrix_client,
            self.fake_room.room_id,
            "Created silence with ID silence1.",
        )
        self.fake_cache.__getitem__.assert_called_once_with(self.fake_alert_event_id)
        self.fake_cache.set.assert_called_once_with(
            "some event id", "fingerprint1", expire=None
        )

    @patch.object(matrix_alertbot.command, "send_text_to_room")
    async def test_ack_with_duration(self, fake_send_text_to_room: Mock) -> None:
        """Tests the callback for InviteMemberEvents"""
        # Tests that the bot attempts to join a room after being invited to it
        fake_cache_dict = {self.fake_alert_event_id: "fingerprint1"}

        self.fake_cache.__getitem__.side_effect = fake_cache_dict.__getitem__

        command = AckAlertCommand(
            self.fake_matrix_client,
            self.fake_cache,
            self.fake_alertmanager_client,
            self.fake_config,
            self.fake_room,
            self.fake_sender,
            self.fake_event_id,
            self.fake_alert_event_id,
            ("1w", "3d"),
        )
        await command.process()

        # Check that we attempted to create silences
        self.fake_alertmanager_client.create_or_update_silence.assert_called_once_with(
            "fingerprint1", self.fake_sender, 864000, force=True
        )
        fake_send_text_to_room.assert_called_once_with(
            self.fake_matrix_client,
            self.fake_room.room_id,
            "Created silence with ID silence1.",
        )
        self.fake_cache.__getitem__.assert_called_once_with(self.fake_alert_event_id)
        self.fake_cache.set.assert_called_once_with(
            "some event id", "fingerprint1", expire=864000
        )

    @patch.object(matrix_alertbot.command, "send_text_to_room")
    async def test_ack_raise_alertmanager_error(
        self, fake_send_text_to_room: Mock
    ) -> None:
        """Tests the callback for InviteMemberEvents"""
        # Tests that the bot attempts to join a room after being invited to it
        fake_cache_dict = {
            self.fake_alert_event_id: "fingerprint1",
        }

        self.fake_cache.__getitem__.side_effect = fake_cache_dict.__getitem__
        self.fake_alertmanager_client.create_or_update_silence.side_effect = (
            create_silence_raise_alertmanager_error
        )

        command = AckAlertCommand(
            self.fake_matrix_client,
            self.fake_cache,
            self.fake_alertmanager_client,
            self.fake_config,
            self.fake_room,
            self.fake_sender,
            self.fake_event_id,
            self.fake_alert_event_id,
        )
        await command.process()

        # Check that we attempted to create silences
        self.fake_alertmanager_client.create_or_update_silence.assert_called_once_with(
            "fingerprint1", self.fake_sender, None, force=True
        )
        fake_send_text_to_room.assert_called_once_with(
            self.fake_matrix_client,
            self.fake_room.room_id,
            "Sorry, I couldn't create silence for alert with fingerprint fingerprint1 because something went wrong with Alertmanager: ",
        )
        self.fake_cache.__getitem__.assert_called_once_with(self.fake_alert_event_id)
        self.fake_cache.set.assert_not_called()

    @patch.object(matrix_alertbot.command, "send_text_to_room")
    async def test_ack_raise_alert_not_found_error(
        self, fake_send_text_to_room: Mock
    ) -> None:
        """Tests the callback for InviteMemberEvents"""
        # Tests that the bot attempts to join a room after being invited to it
        fake_cache_dict = {
            self.fake_alert_event_id: "fingerprint1",
        }

        self.fake_cache.__getitem__.side_effect = fake_cache_dict.__getitem__
        self.fake_alertmanager_client.create_or_update_silence.side_effect = (
            create_silence_raise_alert_not_found_error
        )

        command = AckAlertCommand(
            self.fake_matrix_client,
            self.fake_cache,
            self.fake_alertmanager_client,
            self.fake_config,
            self.fake_room,
            self.fake_sender,
            self.fake_event_id,
            self.fake_alert_event_id,
        )
        await command.process()

        # Check that we attempted to create silences
        self.fake_alertmanager_client.create_or_update_silence.assert_called_once_with(
            "fingerprint1", self.fake_sender, None, force=True
        )
        fake_send_text_to_room.assert_called_once_with(
            self.fake_matrix_client,
            self.fake_room.room_id,
            "Sorry, I couldn't create silence for alert with fingerprint fingerprint1: ",
        )
        self.fake_cache.__getitem__.assert_called_once_with(self.fake_alert_event_id)
        self.fake_cache.set.assert_not_called()

    @patch.object(matrix_alertbot.command, "send_text_to_room")
    async def test_ack_with_invalid_duration(
        self, fake_send_text_to_room: Mock
    ) -> None:
        """Tests the callback for InviteMemberEvents"""
        # Tests that the bot attempts to join a room after being invited to it
        command = AckAlertCommand(
            self.fake_matrix_client,
            self.fake_cache,
            self.fake_alertmanager_client,
            self.fake_config,
            self.fake_room,
            self.fake_sender,
            self.fake_event_id,
            self.fake_alert_event_id,
            ("invalid duration",),
        )

        await command.process()

        # Check that we attempted to create silences
        self.fake_alertmanager_client.create_silence.assert_not_called()
        fake_send_text_to_room.assert_called_once_with(
            self.fake_matrix_client,
            self.fake_room.room_id,
            "I tried really hard, but I can't convert the duration 'invalid duration' to a number of seconds.",
        )
        self.fake_cache.__getitem__.assert_not_called()
        self.fake_cache.get.assert_not_called()
        self.fake_cache.set.assert_not_called()

    @patch.object(matrix_alertbot.command, "send_text_to_room")
    async def test_ack_with_negative_duration(
        self, fake_send_text_to_room: Mock
    ) -> None:
        """Tests the callback for InviteMemberEvents"""
        # Tests that the bot attempts to join a room after being invited to it
        command = AckAlertCommand(
            self.fake_matrix_client,
            self.fake_cache,
            self.fake_alertmanager_client,
            self.fake_config,
            self.fake_room,
            self.fake_sender,
            self.fake_event_id,
            self.fake_alert_event_id,
            ("-1d",),
        )

        await command.process()

        # Check that we attempted to create silences
        self.fake_alertmanager_client.create_silence.assert_not_called()
        fake_send_text_to_room.assert_called_once_with(
            self.fake_matrix_client,
            self.fake_room.room_id,
            "I can't create a silence with a negative duration!",
        )
        self.fake_cache.__getitem__.assert_not_called()
        self.fake_cache.get.assert_not_called()
        self.fake_cache.set.assert_not_called()

    @patch.object(matrix_alertbot.command, "send_text_to_room")
    async def test_ack_with_alert_event_not_found_in_cache(
        self, fake_send_text_to_room: Mock
    ) -> None:
        """Tests the callback for InviteMemberEvents"""
        # Tests that the bot attempts to join a room after being invited to it
        fake_cache_dict: Dict = {}

        self.fake_cache.__getitem__.side_effect = fake_cache_dict.__getitem__
        self.fake_cache.get.side_effect = fake_cache_dict.get

        command = AckAlertCommand(
            self.fake_matrix_client,
            self.fake_cache,
            self.fake_alertmanager_client,
            self.fake_config,
            self.fake_room,
            self.fake_sender,
            self.fake_event_id,
            self.fake_alert_event_id,
        )

        await command.process()

        # Check that we attempted to create silences
        self.fake_alertmanager_client.create_silence.assert_not_called()
        fake_send_text_to_room.assert_not_called()
        self.fake_cache.__getitem__.assert_called_once_with("some alert event id")
        self.fake_cache.get.assert_not_called()
        self.fake_cache.set.assert_not_called()

    @patch.object(matrix_alertbot.command, "send_text_to_room")
    async def test_unack(self, fake_send_text_to_room: Mock) -> None:
        """Tests the callback for InviteMemberEvents"""
        # Tests that the bot attempts to join a room after being invited to it
        fake_cache_dict = {
            self.fake_alert_event_id: "fingerprint1",
            "fingerprint1": "silence1",
        }

        self.fake_cache.__getitem__.side_effect = fake_cache_dict.__getitem__

        command = UnackAlertCommand(
            self.fake_matrix_client,
            self.fake_cache,
            self.fake_alertmanager_client,
            self.fake_config,
            self.fake_room,
            self.fake_sender,
            self.fake_event_id,
            self.fake_alert_event_id,
        )
        await command.process()

        # Check that we attempted to create silences
        self.fake_alertmanager_client.delete_silence.assert_called_once_with("silence1")
        fake_send_text_to_room.assert_called_once_with(
            self.fake_matrix_client,
            self.fake_room.room_id,
            "Removed silence with ID silence1.",
        )
        self.fake_cache.__getitem__.assert_has_calls(
            [call(self.fake_alert_event_id), call("fingerprint1")]
        )

    @patch.object(matrix_alertbot.command, "send_text_to_room")
    async def test_unack_silence_raise_alertmanager_error(
        self, fake_send_text_to_room: Mock
    ) -> None:
        """Tests the callback for InviteMemberEvents"""
        # Tests that the bot attempts to join a room after being invited to it
        fake_cache_dict = {
            self.fake_alert_event_id: "fingerprint1",
            "fingerprint1": "silence1",
        }

        self.fake_cache.__getitem__.side_effect = fake_cache_dict.__getitem__

        command = UnackAlertCommand(
            self.fake_matrix_client,
            self.fake_cache,
            self.fake_alertmanager_client,
            self.fake_config,
            self.fake_room,
            self.fake_sender,
            self.fake_event_id,
            self.fake_alert_event_id,
        )

        self.fake_alertmanager_client.delete_silence.side_effect = (
            delete_silence_raise_alertmanager_error
        )
        await command.process()

        # Check that we attempted to create silences
        self.fake_alertmanager_client.delete_silence.assert_called_once_with("silence1")
        fake_send_text_to_room.assert_called_once_with(
            self.fake_matrix_client,
            self.fake_room.room_id,
            "Sorry, I couldn't remove silence for alert with fingerprint fingerprint1 because something went wrong with Alertmanager: ",
        )
        self.fake_cache.__getitem__.assert_has_calls(
            [call(self.fake_alert_event_id), call("fingerprint1")]
        )

    @patch.object(matrix_alertbot.command, "send_text_to_room")
    async def test_unack_raise_silence_not_found_error(
        self, fake_send_text_to_room: Mock
    ) -> None:
        """Tests the callback for InviteMemberEvents"""
        # Tests that the bot attempts to join a room after being invited to it
        fake_cache_dict = {
            self.fake_alert_event_id: "fingerprint1",
            "fingerprint1": "silence1",
        }

        self.fake_cache.__getitem__.side_effect = fake_cache_dict.__getitem__

        command = UnackAlertCommand(
            self.fake_matrix_client,
            self.fake_cache,
            self.fake_alertmanager_client,
            self.fake_config,
            self.fake_room,
            self.fake_sender,
            self.fake_event_id,
            self.fake_alert_event_id,
        )

        self.fake_alertmanager_client.delete_silence.side_effect = (
            delete_silence_raise_silence_not_found_error
        )
        await command.process()

        # Check that we attempted to create silences
        self.fake_alertmanager_client.delete_silence.assert_called_once_with("silence1")
        fake_send_text_to_room.assert_called_once_with(
            self.fake_matrix_client,
            self.fake_room.room_id,
            "Sorry, I couldn't remove silence for alert with fingerprint fingerprint1: ",
        )
        self.fake_cache.__getitem__.assert_has_calls(
            [call(self.fake_alert_event_id), call("fingerprint1")]
        )

    @patch.object(matrix_alertbot.command, "send_text_to_room")
    async def test_unack_with_event_not_found_in_cache(
        self, fake_send_text_to_room: Mock
    ) -> None:
        """Tests the callback for InviteMemberEvents"""
        # Tests that the bot attempts to join a room after being invited to it
        fake_cache_dict: Dict = {}

        self.fake_cache.__getitem__.side_effect = fake_cache_dict.__getitem__

        command = UnackAlertCommand(
            self.fake_matrix_client,
            self.fake_cache,
            self.fake_alertmanager_client,
            self.fake_config,
            self.fake_room,
            self.fake_sender,
            self.fake_event_id,
            self.fake_alert_event_id,
        )

        await command.process()

        # Check that we attempted to create silences
        self.fake_alertmanager_client.delete_silence.assert_not_called()
        fake_send_text_to_room.assert_not_called()
        self.fake_cache.__getitem__.assert_called_once_with(self.fake_alert_event_id)

    @patch.object(matrix_alertbot.command, "send_text_to_room")
    async def test_unack_with_silence_not_found_in_cache(
        self, fake_send_text_to_room: Mock
    ) -> None:
        """Tests the callback for InviteMemberEvents"""
        # Tests that the bot attempts to join a room after being invited to it
        fake_cache_dict = {self.fake_alert_event_id: "fingerprint1"}

        self.fake_cache.__getitem__.side_effect = fake_cache_dict.__getitem__

        command = UnackAlertCommand(
            self.fake_matrix_client,
            self.fake_cache,
            self.fake_alertmanager_client,
            self.fake_config,
            self.fake_room,
            self.fake_sender,
            self.fake_event_id,
            self.fake_alert_event_id,
        )

        await command.process()

        # Check that we attempted to create silences
        self.fake_alertmanager_client.delete_silence.assert_not_called()
        fake_send_text_to_room.assert_not_called()
        self.fake_cache.__getitem__.assert_has_calls(
            [call(self.fake_alert_event_id), call("fingerprint1")]
        )

    @patch.object(matrix_alertbot.command, "send_text_to_room")
    async def test_help_without_topic(self, fake_send_text_to_room: Mock) -> None:
        """Tests the callback for InviteMemberEvents"""
        # Tests that the bot attempts to join a room after being invited to it

        command = HelpCommand(
            self.fake_matrix_client,
            self.fake_cache,
            self.fake_alertmanager_client,
            self.fake_config,
            self.fake_room,
            self.fake_sender,
            self.fake_event_id,
        )

        await command.process()

        # Check that we attempted to create silences
        fake_send_text_to_room.assert_called_once()
        _, _, text = fake_send_text_to_room.call_args.args
        self.assertIn("help commands", text)

    @patch.object(matrix_alertbot.command, "send_text_to_room")
    async def test_help_with_rules_topic(self, fake_send_text_to_room: Mock) -> None:
        """Tests the callback for InviteMemberEvents"""
        # Tests that the bot attempts to join a room after being invited to it

        command = HelpCommand(
            self.fake_matrix_client,
            self.fake_cache,
            self.fake_alertmanager_client,
            self.fake_config,
            self.fake_room,
            self.fake_sender,
            self.fake_event_id,
            ("rules",),
        )

        await command.process()

        # Check that we attempted to create silences
        fake_send_text_to_room.assert_called_once()
        _, _, text = fake_send_text_to_room.call_args.args
        self.assertIn("rules!", text)

    @patch.object(matrix_alertbot.command, "send_text_to_room")
    async def test_help_with_commands_topic(self, fake_send_text_to_room: Mock) -> None:
        """Tests the callback for InviteMemberEvents"""
        # Tests that the bot attempts to join a room after being invited to it

        command = HelpCommand(
            self.fake_matrix_client,
            self.fake_cache,
            self.fake_alertmanager_client,
            self.fake_config,
            self.fake_room,
            self.fake_sender,
            self.fake_event_id,
            ("commands",),
        )

        await command.process()

        # Check that we attempted to create silences
        fake_send_text_to_room.assert_called_once()
        _, _, text = fake_send_text_to_room.call_args.args
        self.assertIn("Available commands", text)

    @patch.object(matrix_alertbot.command, "send_text_to_room")
    async def test_help_with_unknown_topic(self, fake_send_text_to_room: Mock) -> None:
        """Tests the callback for InviteMemberEvents"""
        # Tests that the bot attempts to join a room after being invited to it

        command = HelpCommand(
            self.fake_matrix_client,
            self.fake_cache,
            self.fake_alertmanager_client,
            self.fake_config,
            self.fake_room,
            self.fake_sender,
            self.fake_event_id,
            ("unknown",),
        )

        await command.process()

        # Check that we attempted to create silences
        fake_send_text_to_room.assert_called_once()
        _, _, text = fake_send_text_to_room.call_args.args
        self.assertEqual("Unknown help topic!", text)

    @patch.object(matrix_alertbot.command, "send_text_to_room")
    async def test_unknown_command(self, fake_send_text_to_room: Mock) -> None:
        """Tests the callback for InviteMemberEvents"""
        # Tests that the bot attempts to join a room after being invited to it

        command = UnknownCommand(
            self.fake_matrix_client,
            self.fake_cache,
            self.fake_alertmanager_client,
            self.fake_config,
            self.fake_room,
            self.fake_sender,
            self.fake_event_id,
        )

        await command.process()

        # Check that we attempted to create silences
        fake_send_text_to_room.assert_called_once_with(
            self.fake_matrix_client,
            self.fake_room.room_id,
            "Unknown command. Try the 'help' command for more information.",
        )


if __name__ == "__main__":
    unittest.main()



# This file holds custom error types that you can define for your application.


class MatrixAlertbotError(Exception):
    pass


class ConfigError(MatrixAlertbotError):
    """An error encountered during reading the config file."""

    pass


class ParseConfigError(ConfigError):
    """An error encountered when config file cannot be parsed."""

    pass


class InvalidConfigError(ParseConfigError):
    """An error encountered when a config key is not valid."""

    pass


class RequiredConfigKeyError(ConfigError):
    """An error encountered when a required config key is missing."""

    pass


class AlertmanagerError(MatrixAlertbotError):
    """An error encountered with Alertmanager."""

    pass


class AlertNotFoundError(AlertmanagerError):
    """An error encountered when an alert cannot be found in Alertmanager."""

    pass


class SilenceNotFoundError(AlertmanagerError):
    """An error encountered when a silence cannot be found in Alertmanager."""

    pass


class SilenceExpiredError(AlertmanagerError):
    """An error encountered when a silence is already expired in Alertmanager."""

    pass


class SilenceExtendError(AlertmanagerError):
    """An error encountered when a silence cannot be extended."""

    pass


class AlertmanagerServerError(AlertmanagerError):
    """An error encountered with Alertmanager server."""

    pass
